#include "CppGenerator.h"

#include <fstream>
#include <sstream>
#include <iostream>
#include <algorithm>
#include <sys/stat.h>



CppGenerator::CppGenerator()
{
    //ctor
}

void CppGenerator::AddClass(ClassHead *head, pnNet *_objectNet, pnClassContent *content)
{
   pnClassCpp *c = new pnClassCpp(head, _objectNet, content);
   this->classes.push_back((pnClassCpp*)c);
}

pnClass *CppGenerator::CreateClass(ClassHead *head, pnNet *_objectNet, pnClassContent *content)
{
    return new pnClassCpp(head, _objectNet, content);
}

CppGenerator::~CppGenerator()
{
    list<pnClassCpp*>::iterator it_var, it_end;
    it_var = classes.begin();
    it_end = classes.end();

    for(; it_var != it_end; it_var++)
    {
        delete *it_var;
    }
}

void pnClassCpp::addMethod(pnClassMethod* net)
{
    this->methods.push_back(net);
}

void pnClassCpp::addConstructor(pnClassMethod* net)
{
    this->constructors.push_back(net);
}

void pnClassCpp::addSyncPort(pnClassSync* sync)
{
    this->syncPorts.push_back(sync);
}

void pnClassCpp::Print()
{
    pnClass::Print();

    list<pnClassMethod*>::iterator it_var, it_end;
    it_var = methods.begin();
    it_end = methods.end();

    for(; it_var != it_end; it_var++)
    {
        cout << "\t method: ";
        (*it_var)->Print();
    }
}

list<string> pnClassCpp::GetRequiredClasses()
{
    return this->objectNet->GetMetaClasses();
}

pnClassCpp::~pnClassCpp()
{
    list<pnClassMethod*>::iterator it_var, it_end;
    it_var = this->methods.begin();
    it_end = this->methods.end();

    for(; it_var != it_end; it_var++)
    {
        delete *it_var;
    }
    cout << "deleting class " << this->id << endl;
}

/** @brief CreateMultisetItemSymbol
  *
  *
  */
SymbolLiteral* CppGenerator::CreateMultisetItemSymbol(string value)
{
    return new SymbolLiteralCpp(value);
}

/** @brief CreateMultisetItemString
  *
  *
  */
StringLiteral* CppGenerator::CreateMultisetItemString(string value)
{
    return new StringLiteralCpp(value);
}

/** @brief CreateMultisetItemChar
  *
  *
  */
CharLiteral* CppGenerator::CreateMultisetItemChar(char value)
{
    return new CharLiteralCpp(value);
}

/** @brief CreateMultisetItemInt
  *
  *
  */
IntegerLiteral* CppGenerator::CreateMultisetItemInt(int value)
{
    return new IntegerLiteralCpp(value);
}

/** @brief CreateMultisetItemVar
  *
  *
  */
pnSetItem* CppGenerator::CreateMultisetItemVar(string varName)
{
    if(varName == "true")
        return new pnMultisetItemLiteralCpp(new BoolLiteralCpp(true));
    else if (varName == "false")
        return new pnMultisetItemLiteralCpp(new BoolLiteralCpp(true));
    else if (varName == "nil")
        return new pnMultisetItemLiteralCpp(new NilLiteralCpp());
    else
        return new pnMultisetItemVarCpp(varName);
}

/** @brief CreateMultisetItemLiteral
  *
  *
  */
pnSetItem* CppGenerator::CreateMultisetItemLiteral(Literal* literal)
{
    return new pnMultisetItemLiteralCpp(literal);
}

/** @brief CreateMultisetItemList
  *
  *
  */
pnMultisetItemList* CppGenerator::CreateMultisetItemList()
{
    return new pnMultisetItemListCpp();
}

/** @brief CreateKeysMessage
  *
  *
  */
pnMessageKeys* CppGenerator::CreateKeysMessage()
{
    return new KeysMessageCpp();
}

/** @brief CreateBinaryMessage
  *
  *
  */
pnMessageBinary* CppGenerator::CreateBinaryMessage(string msg, pnOperand* op, pnMessageUnary *opChain)
{
    return new BinaryMessageCpp(msg, op, opChain);
}

/** @brief CreateUnaryMessage
  *
  */
pnMessageUnary* CppGenerator::CreateUnaryMessage(string msg, pnMessageUnary *nextMsg)
{
    return new UnaryMessageCpp(msg, nextMsg);
}

/** @brief CreateOperandToken
  *
  *
  */
pnOperandLiteral* CppGenerator::CreateOperandToken(Literal *token)
{
    return new OperandTokenCpp(token);
}

/** @brief CreateOperandVar
  *
  *
  */
pnOperandVar* CppGenerator::CreateOperandVar(string varName)
{
    return new OperandVarCpp(varName);
}



/** @brief GenerateCode
  *
  *
  */
void CppGenerator::GenerateCode(string dir, string programName)
{
    list<pnClassCpp*>::iterator it_var, it_end;
    it_var = classes.begin();
    it_end = classes.end();

    for(; it_var != it_end; it_var++)
    {
        GeneratePNHeader(*it_var, dir);
        GeneratePNClass(*it_var, dir);
    }
    GenerateMain(dir);

    list<string> headers;
    headers.push_back("PetriNet.h");
    headers.push_back("Token.h");
    headers.push_back("PN_Transition.h");
    headers.push_back("PN_Place.h");
    headers.push_back("PNsimulation.h");
    headers.push_back("PN_Multiset.h");
    headers.push_back("Bindings.h");
    headers.push_back("PN_Condition.h");
    headers.push_back("PN_Exception.h");
    headers.push_back("PN_Smalltalk.h");

    ifstream  src;
    ofstream  dst;

    src.open("./PN_Sim/libPN_Sim.a", std::ios::binary);
    dst.open((dir + "/libPN_Sim.a").c_str(), std::ios::binary);
    dst << src.rdbuf();
    dst.close();
    src.close();

    list<string>::iterator it_var2, it_end2;
    it_var2 = headers.begin();
    it_end2 = headers.end();

    for(; it_var2 != it_end2; it_var2++)
    {
        src.open(("./PN_Sim/" + *it_var2).c_str(), std::ios::binary);
        dst.open((dir + "/" + *it_var2).c_str(), std::ios::binary);
        dst << src.rdbuf();
        dst.close();
        src.close();
    }


    GenerateMake(dir, programName);
}

void CppGenerator::GenerateMain(string dir)
{
    ostringstream text(ostringstream::app);
    ofstream classFile;
    string cpp_filename = dir + "/main.cpp";
    classFile.open (cpp_filename.c_str(), ios::trunc);
    if(classFile.fail())
    {
        cout << "failed to create file " << cpp_filename << endl ;
        classFile.close();
        exit(1);
    }

    text << "#include \"PNsimulation.h\"" << endl;
    text << "#include \"" << this->startClass << ".h\""<< endl << endl;

    text << "int main()" << endl;
    text << "{" << endl;
    text << "\t" << this->startClass << " *startClass = new " << this->startClass << "();" << endl;
    text << "\tPNsimulation::Instance()->Run();" << endl;
    text << "\tdelete startClass;" << endl;
    text << "\treturn 0;" << endl;
    text << "}" << endl;

    classFile << text.str();
    classFile.close();
}
void CppGenerator::GenerateMake(string dir, string programName)
{
    ostringstream text(ostringstream::app);
    ofstream classFile;
    string cpp_filename = dir + "/Makefile";
    classFile.open (cpp_filename.c_str(), ios::trunc);
    if(classFile.fail())
    {
        cout << "failed to create file " << cpp_filename << endl ;
        classFile.close();
        exit(1);
    }

    text << "PROG_NAME = " << programName << endl<< endl;

    text << "CXX = g++" << endl;
    text << "OBJDIR = obj" << endl;
    text << "CFLAGS = -Wall -fexceptions -g -O2" << endl;
    text << "SIM_LIB = libPN_Sim.a" << endl << endl;

    text << "OBJ = $(OBJDIR)/main.o";

    list<pnClassCpp*>::iterator it_var, it_end;
    it_var = classes.begin();
    it_end = classes.end();
    for(; it_var != it_end; it_var++)
    {
        text <<  " $(OBJDIR)/" << (*it_var)->id << ".o";
    }
    text << endl << endl;

    text << "all: build" << endl << endl;

    text << "before_build:" << endl;
	text << "\ttest -d $(OBJDIR) || mkdir $(OBJDIR)" << endl<< endl;

    text << "build: before_build $(OBJ)" << endl;
	text << "\t$(CXX) -o $(PROG_NAME) $(OBJ) $(SIM_LIB)" << endl<< endl;

    text << "clean:\n\trm -rf $(OBJDIR)" << endl<< endl;

    it_var = classes.begin();
    it_end = classes.end();

    for(; it_var != it_end; it_var++)
    {
       text << "$(OBJDIR)/" << (*it_var)->id << ".o: " << (*it_var)->id << ".cpp" << endl;
        text << "\t$(CXX) $(CFLAGS) -c -o $(OBJDIR)/" << (*it_var)->id << ".o " << (*it_var)->id << ".cpp" << endl<< endl;
    }

    text << "$(OBJDIR)/main.o: main.cpp" << endl;
	text << "\t$(CXX) $(CFLAGS) -c -o $(OBJDIR)/main.o main.cpp" << endl;

    classFile << text.str();
    classFile.close();
}

/** @brief GeneratePNHeader
 *
 *
 */
void CppGenerator::GeneratePNHeader(pnClassCpp *c, string dir)
{
    ostringstream text(ostringstream::app);
    ofstream headerFile;
    mkdir(dir.c_str(), S_IRWXU | S_IRWXG | S_IRWXO | S_IXOTH);
    string h_filename =  dir + "/" + c->id + ".h";
    headerFile.open (h_filename.c_str(), ios::trunc);
    if(headerFile.fail())
    {
        cout << "failed to create file " << h_filename << endl ;
        headerFile.close();
        exit(1);
    }


    string headerConst = c->id;
    transform(headerConst.begin(), headerConst.end(), headerConst.begin(), ::toupper);

/// file header
    text << "#ifndef "<< headerConst <<"_H" << endl;
    text << "#define "<< headerConst <<"_H" << endl << endl;
    text << "#include \"PetriNet.h\""<< endl;

    list<string> dependencies(c->GetRequiredClasses());
    list<string>::iterator it_var_string, it_end_string;
    it_var_string = dependencies.begin();
    it_end_string = dependencies.end();

    for(; it_var_string != it_end_string; it_var_string++)
    {
        text << "#include \"" << *it_var_string << ".h\"" << endl;
    }

/// Class definition
    text << "class "<< c->id << ": public " << c->parentName << endl;   /// class def
    text << "{\npublic: \n";
    text << "\t"<< c->id << "();" << endl;            /// constructor
    text << "\tvirtual ~" << c->id << "();\n";        /// destructor
    text << "\tstatic PN_MetaClass *GetInstantiator();\n";
    if(!c->syncPorts.empty())
    {
        text << "\tvirtual Token *SimulateMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);\n";
    }
    if(!(c->methods.empty() && c->syncPorts.empty()))
    {
        text << "\tvirtual Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);\n";
    }
    text << "\tPN_Place *GetPlace(string placeName);\n";
    text << "\tstring AsText(){return \""<< c->id <<"\";}\n";
    text << "private:\n";

    ///--------------------- places -------------------------------------
    if(c->objectNet != NULL)
    {
        list<pnPlace*>::iterator it_var, it_end;
        it_var = c->objectNet->places.begin();
        it_end = c->objectNet->places.end();

        for(; it_var != it_end; it_var++)
        {
            text << "\tPN_Place *" << (*it_var)->name << ";\n";    ///place def
        }
    }

    ///-------------------- sync ports -----------------------------

    list<pnClassSync*>::iterator it_var_port, it_end_port;
    it_var_port = c->syncPorts.begin();
    it_end_port = c->syncPorts.end();

    for(; it_var_port != it_end_port; it_var_port++)
    {
        text << "\tPN_SyncPort *" << (*it_var_port)->label->messageTarget << ";\n";
    }


    text << "};\n\n";

/// instantiator MetaClass
    text << "class PN_MetaClass_"<< c->id << ": public PN_MetaClass\n"; /// metaclass def
    text << "{\nprotected: \n";
    text << "\tPNToken *GetNewInstance() {return new PNToken(new "<< c->id <<"());}\n";
    text << "};\n\n";

/// methods classes

    list<pnClassMethod*>::iterator it_var, it_end;
    it_var = c->methods.begin();
    it_end = c->methods.end();

    for(; it_var != it_end; it_var++)
    {
        GenerateMethodHeaderCpp(text, *it_var, c->id);
    }

/// footer

    text << "#endif // " << headerConst << "_H" << endl;

    headerFile << text.str();
    headerFile.close();
}

/** @brief GeneratePNClass
  *
  *
  */
void CppGenerator::GeneratePNClass(pnClassCpp *c, string dir)
{
    ostringstream text(ostringstream::app);
    ofstream classFile;
    string cpp_filename = dir + "/" + c->id +".cpp";
    classFile.open (cpp_filename.c_str(), ios::trunc);
    if(classFile.fail())
    {
        cout << "failed to create file " << cpp_filename << endl ;
        classFile.close();
        exit(1);
    }

    text << "#include \""<< c->id <<".h\"\n\n";

/// Get instantiator
    text << "PN_MetaClass *"<< c->id <<"::GetInstantiator()\n";
    text << "{\n";
    text << "\treturn new PN_MetaClass_"<< c->id <<"();\n";
    text << "}\n\n";

/// class constructor
    text << c->id << "::"<< c->id <<"()\n";
    text << "{\n";

    if(c->objectNet != NULL)
    {
        text << "\t// ------------------places init----------------\n";

        list<pnPlace*>::iterator it_var, it_end;
        it_var = c->objectNet->places.begin();
        it_end = c->objectNet->places.end();

        for(; it_var != it_end; it_var++)
        {
            GeneratePlaceInitCpp(text, *it_var);
        }

        text << "\t// ----------------transitions init---------------\n";

        list<pnTransition*>::iterator it_var2, it_end2;
        it_var2 = c->objectNet->transitions.begin();
        it_end2 = c->objectNet->transitions.end();

        for(; it_var2 != it_end2; it_var2++)
        {
            GenerateTransitionInitCpp(text, *it_var2);
        }

        text << "\t//-------------------- sync ports -----------------------------\n";

        list<pnClassSync*>::iterator it_var_port, it_end_port;
        it_var_port = c->syncPorts.begin();
        it_end_port = c->syncPorts.end();

        for(; it_var_port != it_end_port; it_var_port++)
        {
            GenerateSyncPortInitCpp(text, *it_var_port);
        }
    }
    text << "}\n\n";

/// class destructor
    text << c->id <<"::~"<< c->id <<"()\n";
    text << "{\n";

    if(c->objectNet != NULL)                            /// deallocate places
    {
        list<pnPlace*>::iterator it_var, it_end;
        it_var = c->objectNet->places.begin();
        it_end = c->objectNet->places.end();

        for(; it_var != it_end; it_var++)
        {
            text << "\t" << (*it_var)->name <<"->Clear();\n";
            text << "\tdelete " << (*it_var)->name << ";\n";    ///place free
        }
    }
    text << "}\n\n";

/// GetPLace
    text << "PN_Place *" << c->id << "::GetPlace(string placeName)\n";
    text << "{\n";

    if(c->objectNet != NULL)
    {
        list<pnPlace*>::iterator it_var, it_end;
        it_var = c->objectNet->places.begin();
        it_end = c->objectNet->places.end();

        text << "\t";

        for(; it_var != it_end; it_var++)
        {
            text << "if (placeName == \""<< (*it_var)->name <<"\")\n";

            /// we must avoid place named "return" => subtitude to "_return"
            if((*it_var)->name == "return")
            {
                text << "\t\treturn _return;\n";
            }
            else
            {
                text << "\t\treturn "<< (*it_var)->name<< ";\n";
            }

            text << "\telse ";
        }
        text << "\n\t\treturn NULL;\n";
    }
    text << "}\n\n";


    GenerateSendMessage(text, c);
    GenerateSimulateMessage(text, c);

/// methods
    list<pnClassMethod*>::iterator it_var, it_end;
    it_var = c->methods.begin();
    it_end = c->methods.end();

    for(; it_var != it_end; it_var++)
    {
        GenerateMethodContentCpp(text, *it_var, c->id);
    }
    classFile << text.str();
    classFile.close();
}

void GenerateSendMessage(ostringstream& text, pnClassCpp *c)
{
    if (c->methods.empty() && c->syncPorts.empty())
    { return; }

    text << "Token* " << c->id << "::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)\n";
    text << "{\n";

 /// generate sync port effects
    list<pnClassSync*>::iterator it_var_port, it_end_port;
    it_var_port = c->syncPorts.begin();
    it_end_port = c->syncPorts.end();

    for(; it_var_port != it_end_port; it_var_port++)
    {
        text << "\tif(message == \""<< (*it_var_port)->label->messageOriginal << "\")\n";
        text << "\t{\n";
        text << "\t\treturn new BoolToken("<< (*it_var_port)->label->messageTarget <<"->Execute(args, binds));\n";
        text << "\t}\n";
    }


    /// generate method invokers
    list<pnClassMethod*>::iterator it_var_method, it_end_method;
    it_var_method = c->methods.begin();
    it_end_method = c->methods.end();

    for(; it_var_method != it_end_method; it_var_method++)
    {
        int argsCount = (*it_var_method)->label->operands.size();

        text <<"\tif(message == \""<< (*it_var_method)->label->messageOriginal<<"\" )\n";
        text << "\t{\n";

        text << "\t\tPN_Operand *op;\n";

        text << "\t\tif (args.size() < "<< argsCount<<")\n";
        text << "\t\t{\n";
        text << "\t\t\tthrow new PN_NotEnoughArguments(\"" << c->id << "\", \"" <<(*it_var_method)->label->messageOriginal << "\", "<< argsCount <<", args.size());\n";
        text << "\t\t}\n";

        for(int i = 1; i <= argsCount; i++)
        {
            text << "\t\top = args.front();\n";
            text << "\t\targs.pop_front();\n";
            text << "\t\tToken *arg" << i <<" = op->Resolve(binds);\n";
            text << "\t\tif (arg"<< i <<" == NULL)\n";
            text << "\t\t{\n";
            text << "\t\t\treturn NULL;\n";
            text << "\t\t}\n";
        }
        text << "\t\tthrow new PN_MethodForkException(new " << c->id << '_'<< (*it_var_method)->label->messageTarget << "(this";
        for(int i = 1; i <= argsCount; i++)
        {
            text << ", arg" << i;
        }
        text <<"));\n";
        text << "\t}\n";
    }
    text << "\treturn "<< c->parentName << "::SendMessage(message, args, binds);\n";
    text << "}\n\n";

}

void GenerateSimulateMessage(ostringstream& text, pnClassCpp *c)
{
    if (c->syncPorts.empty())
    { return; }

    text << "Token* " << c->id << "::SimulateMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)\n";
    text << "{\n";

 /// generate sync port effects
    list<pnClassSync*>::iterator it_var_port, it_end_port;
    it_var_port = c->syncPorts.begin();
    it_end_port = c->syncPorts.end();

    for(; it_var_port != it_end_port; it_var_port++)
    {
        text << "\tif(message == \""<< (*it_var_port)->label->messageOriginal << "\")\n";
        text << "\t{\n";
        text << "\t\treturn new BoolToken("<< (*it_var_port)->label->messageTarget <<"->TryBind(args, binds));\n";
        text << "\t}\n";
    }
    text << "\treturn "<< c->parentName << "::SimulateMessage(message, args, binds);\n";
    text << "}\n\n";

}

void GeneratePlaceInitCpp(ostringstream& text, pnPlace* place)
{
    string placeInitContent = "";
    string placeName;
    if(place->initContent != NULL)
    {
        placeInitContent = GenerateMultisetCpp(text, place->initContent, place->name);
        placeName = placeInitContent;
        placeInitContent.append(", ");
    }

    /// we must avoid place named "return" => subtitude to "_return"
    text << "\t";
    if (place->name == "return")
    {
        text << "_return";
        text << " = new PN_Place_Return(this);\n\n";
    }
    else
    {
        text << place->name;
        text << " = new PN_Place(";
        text << placeInitContent;
        text << "\"" << place->name << "\");\n\n";
    }
    if(place->initContent != NULL)
    {
        text << "\tdelete " << placeName << ";" << endl;
    }
}

void GenerateConds(ostringstream& text, pnCondition *precond, pnCondition *condition, pnCondition *postcond, string elemName)
{
    /// preconds
    if(precond != NULL)
    {
        string listName = elemName;
        listName.append("_precond");
        text << "\t// ---------preconds-------\n";
        text << "\tlist<PN_Condition *> " << listName << ";\n\n";

        GenerateConditionInitCpp(text, precond, listName);

        text << "\t" << elemName << "->SetPreconds("<< listName <<");\n\n";
    }

    /// conds
    if(condition != NULL)
    {
        string listName = elemName;
        listName.append("_cond");
        text << "\t// -----------conds-------\n";
        text << "\tlist<PN_Condition *> " << listName << ";\n\n";

        GenerateConditionInitCpp(text, condition, listName);

        text << "\t" << elemName << "->SetConds("<< listName <<");\n\n";
    }

    /// postconds
    if(postcond != NULL)
    {
        string listName = elemName;
        listName.append("_postcond");
        text << "\t// ---------postconds-------\n";
        text << "\tlist<PN_Condition *> " << listName << ";\n\n";

        GenerateConditionInitCpp(text, postcond, listName);

        text << "\t" << elemName << "->SetPostconds("<< listName <<");\n\n";
    }
}

void GenerateGuard(ostringstream& text, pnGuard *guard, string elemName)
{
    /// guard

    if(guard != NULL)
    {

        text << "\t// --------- guard -------\n";

        int part_guard = 1;

        string owner = elemName;

        list<pnExprAssignmentTree*>::iterator it_var_g, it_end_g;
        it_var_g = guard->exprList.begin();
        it_end_g = guard->exprList.end();

        for (; it_var_g != it_end_g; it_var_g++, part_guard++)
        {
            ostringstream guardName;
            guardName << elemName << "_guard_" << part_guard;
            string gName = guardName.str();

            text << "\tPN_Guard *"<< gName << " = new PN_Guard();" << endl;

            list<pnExprAssignment*> exprsSimpleGuard;
            (*it_var_g)->Simplify(&exprsSimpleGuard, "");

            list<pnExprAssignment*>::iterator it_var, it_end;
            it_var = exprsSimpleGuard.begin();
            it_end = exprsSimpleGuard.end();

            for (int i = 1; it_var != it_end; it_var++, i++)
            {
                ostringstream exprName ; exprName << gName << "_expr_" << i;

                string eName = exprName.str();

                text << "\tPN_Expression *" << eName << ";\n";

                if((*it_var)->msg != NULL)
                {
                    (*it_var)->msg->GenerateSetAction(text, eName, (*it_var)->target);
                }
                else
                {
                    text << "\t" << eName << " = new PN_Expression(NULL, " << (*it_var)->target->GenerateNewInstance() << ");\n";
                }

                text << "\t" << gName <<"->AddExpr(" << eName << ");" << endl;
            }


            text << "\t" << owner << "->SetGuard(" << gName << ");" << endl << endl;
            owner = gName;
        }
    }
}

void GenerateTransitionInitCpp(ostringstream& text, pnTransition* trans)
{
    text << "//-----------------transition " << trans->name << " init-----------------\n";
    text << "\tPN_Transition *"<< trans->name <<" = new PN_Transition(\""<< trans->name <<"\");\n";
    text << "\tthis->RegisterTransition(" << trans->name << ");\n\n";

    GenerateConds(text, trans->precond, trans->condition, trans->postcond, trans->name);

    /// guard

    GenerateGuard(text, trans->guard, trans->name);

    ///action
    if(trans->action != NULL)
    {
        text << "\t// ---------action-------\n";

        list<pnExprAssignment*>::iterator it_var, it_end;
        it_var = trans->action->exprsSimple.begin();
        it_end = trans->action->exprsSimple.end();

        for (int i = 1; it_var != it_end; it_var++, i++)
        {
            ostringstream actionName ; actionName << trans->name << "_action_" << i;

            string aName = actionName.str();

            text << "\tPN_Expression *" << aName << ";\n";

            if((*it_var)->msg != NULL)
            {
                (*it_var)->msg->GenerateSetAction(text, aName, (*it_var)->target);
            }
            else
            {
                text << "\t" << aName << " = new PN_Expression(NULL, " << (*it_var)->target->GenerateNewInstance() << ");\n";
            }

            list<string> :: iterator it_varv, it_endv;
            it_varv = (*it_var)->resultVars.begin();
            it_endv = (*it_var)->resultVars.end();

            for (; it_varv != it_endv; it_varv++)
            {
                text << "\t" << aName << "->AddResultVar(\"" << *it_varv << "\");\n";
            }
            text << "\t" << trans->name <<"->AddAction(" << aName << ");\n\n";
        }
    }
}

void GenerateSyncPortInitCpp(ostringstream& text, pnClassSync *port)
{
    text << "//------------------------ port "<< port->label->messageOriginal << " -----------------------------\n";
    if (port->label->operands.size() > 0)
    {
        string portArgsName = port->label->messageTarget +"_args";
        text << "\tlist<string> "<< portArgsName <<" = list<string>();" << endl;

        list<string> :: iterator it_var, it_end;
        it_var = port->label->operands.begin();
        it_end = port->label->operands.end();

        for (; it_var != it_end; it_var++)
        {
            text << "\t" << portArgsName << ".push_back(\""<< *it_var << "\");" << endl;
        }

        text << "\t" << port->label->messageTarget << " = new PN_SyncPort(" << portArgsName << ", \""<< port->label->messageOriginal << "\");" << endl;
    }
    else
    {
        text << "\t" << port->label->messageTarget << " = new PN_SyncPort(\""<< port->label->messageOriginal << "\");" << endl;
    }

    text << endl;

    GenerateConds(text, port->precond, port->condition, port->postcond, port->label->messageTarget);

    GenerateGuard(text, port->guard, port->label->messageTarget);
}

void GenerateConditionInitCpp(ostringstream& text, pnCondition* cond, string listName)
{
    if(cond == NULL)
    {
        return;
    }

    string varName = listName;
    varName.append(cond->GetTarget());
    varName = GenerateMultisetCpp(text, cond->GetMultiset(), varName);

    string targetPlace = cond->GetTarget();
    if (targetPlace == "return")
        targetPlace = "_return";

    if (cond->UsePlaceSearch())
    {
        ostringstream targetRef;
        targetRef << "this->GetPlace(\""<< targetPlace << "\")";
        targetPlace = targetRef.str();
    }

    text << "\t" << listName <<".push_back(new PN_Condition(" << targetPlace << ", "<< varName <<"));\n";

    GenerateConditionInitCpp(text, cond->GetNextCondition(), listName);
}

string GenerateMultisetCpp(ostringstream& text, pnMultiset* set, string varPrefix)
{
    string varName = varPrefix.append("_Multiset");
    text << "\tPN_Multiset *" << varName <<  " = new PN_Multiset();\n";

    list<pnMultisetItem*>::iterator it_var, it_end;
    it_var = set->mItems.begin();
    it_end = set->mItems.end();

    for(; it_var != it_end; it_var++)
    {
        (*it_var)->GenerateAddToMultiSet(text, varName);
    }
    return varName;
}

string GenerateIndent(int level)
{
    string result;
    for (int i = 0; i < level; i++)
    {
        result.append("\t");
    }
    return result;
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void StringLiteralCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    text << GenerateIndent(level) << varName << "->AddItem(new PN_MultiItemString(" << quantity << ", \"" << this->Value << "\"));\n";
}

string StringLiteralCpp::GenerateNewInstance()
{
    ostringstream text;
    text << "new StringToken(\"" << this->Value << "\")";
    return text.str();
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void SymbolLiteralCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    text << GenerateIndent(level) << varName << "->AddItem(new PN_MultiItemSymbol(" << quantity << ", \"" << this->Value << "\"));\n";
}

string SymbolLiteralCpp::GenerateNewInstance()
{
    ostringstream text;
    text << "new SymbolToken(\"" << this->Value << "\")";
    return text.str();
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void CharLiteralCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    text << GenerateIndent(level) << varName << "->AddItem(new PN_MultiItemChar(" << quantity << ", '" << this->Value << "'));\n";
}

string CharLiteralCpp::GenerateNewInstance()
{
    ostringstream text;
    text << "new CharToken('" << this->Value << "')";
    return text.str();
}

string BoolLiteralCpp::ToString()
{
    return this->Value? "true" : "false";
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void BoolLiteralCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    text << GenerateIndent(level) << varName << "->AddItem(new PN_MultiItemBool(" << quantity << ", " << this->ToString() << "));\n";
}



string BoolLiteralCpp::GenerateNewInstance()
{
    ostringstream text;
    text << "new BoolToken(\"" << this->Value << "\")";
    return text.str();
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void NilLiteralCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    text << GenerateIndent(level) << varName << "->AddItem(new PN_MultiItemNil(" << quantity << "));\n";
}

string NilLiteralCpp::GenerateNewInstance()
{
    ostringstream text;
    text << "new NilToken()";
    return text.str();
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void IntegerLiteralCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    text << GenerateIndent(level) << varName << "->AddItem(new PN_MultiItemInt(" << quantity << ", " << this->Value << "));\n";
}

string IntegerLiteralCpp::GenerateNewInstance()
{
    ostringstream text;
    text << "new IntToken(" << this->Value << ")";
    return text.str();
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void pnMultisetItemListCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    ostringstream listName;
    listName << "multiList" << level;

    text << GenerateIndent(level) << "{\n";
    text << GenerateIndent(level+1) << "PN_MultiItemList *" << listName.str() << " = new PN_MultiItemList(" << quantity << ");\n";
    list<pnSetItem*>::iterator it_var, it_end;
    it_var = this->mList.begin();
    it_end = this->mList.end();

    for (; it_var != it_end; it_var++)
    {
        (*it_var)->GenerateAddToMultiSet(text, listName.str() , "1", level + 1);
    }
    text << GenerateIndent(level + 1) << varName << "->AddItem(" << listName.str() << ");\n";
    text << GenerateIndent(level) << "}\n";
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void pnMultisetItemLiteralCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    this->_literal->GenerateAddToMultiSet(text, varName, quantity, level);
}

/** @brief GenerateAddToMultiSet
  *
  * @todo: document this function
  */
void pnMultisetItemVarCpp::GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level)
{
    text << GenerateIndent(level) << varName << "->AddItem(new PN_MultiItemVariable(" << quantity << ", \"" << this->varName << "\"));\n";
}




/** @brief GenerateSetAction
  *
  * @todo: document this function
  */
void UnaryMessageCpp::GenerateSetAction(ostringstream& text, string actionName, pnOperand *target)
{
    text << "\t" << actionName << " = new PN_Expression(new PN_MessageUnary(\""<< this->message <<"\"), ";
    text << target->GenerateNewInstance();
    text << ");\n";
}

/** @brief GenerateSetAction
  *
  * @todo: document this function
  */
void KeysMessageCpp::GenerateSetAction(ostringstream& text, string actionName, pnOperand *target)
{
    string keyMsgName (actionName);
    keyMsgName.append("_key");
    text << "\tPN_MessageKeys *" << keyMsgName << " = new PN_MessageKeys(\""<< this->message <<"\");\n";

    list<pnMessageBinary*>::iterator it_var, it_end;
    it_var = this->keys.begin();
    it_end = this->keys.end();

    for (; it_var != it_end; it_var++)
    {
        text << "\t" << keyMsgName << "->keys.push_back(" << (*it_var)->operand->GenerateNewInstance() << ");\n";
    }

    text << "\t" << actionName << " = new PN_Expression("<< keyMsgName << ", ";
    text << target->GenerateNewInstance();
    text << ");\n";

}

/** @brief GenerateSetAction
  *
  * @todo: document this function
  */
void BinaryMessageCpp::GenerateSetAction(ostringstream& text, string actionName, pnOperand *target)
{
    text << "\t" << actionName << " = new PN_Expression(new PN_MessageBin(\""<< this->message <<"\", "<<  this->operand->GenerateNewInstance() <<"), ";
    text << target->GenerateNewInstance();
    text << ");\n";
}


void GenerateMethodHeaderCpp(ostringstream& text, pnClassMethod *m, string className)
{
    text << endl;
    string methodClassName = className + "_" + m->label->messageTarget;
    text << "class "<< methodClassName <<" : public PN_Method_Net" << endl;
    text << "{" << endl;
    text << "public:" << endl;
    text << "\t"<< methodClassName <<" (PN_Class *parentObj";

    /// constructor parameters
    list<string>::iterator it_var, it_end;
    it_var = m->label->operands.begin();
    it_end = m->label->operands.end();

    for(; it_var != it_end; it_var++)
    {
        text << ", Token *" << *it_var << "_Operand";
    }

    text << ");" << endl;
    text << "\t" << "virtual ~"<<methodClassName<<"();" << endl;
    text << "private:" << endl;

 ///------------- places --------------------------------------
    if(m->methodNet != NULL)
    {
        list<pnPlace*>::iterator it_var, it_end;
        it_var = m->methodNet->places.begin();
        it_end = m->methodNet->places.end();

        for(; it_var != it_end; it_var++)
        {
            if((*it_var)->name == "return")
            {
                text << "\tPN_Place_Return *_return;\n";
            }
            else
            {
                text << "\tPN_Place *" << (*it_var)->name << ";\n";    ///place def
            }
        }
    }

    text << "};" << endl;
}

void GenerateMethodContentCpp(ostringstream& text, pnClassMethod *m, string className)
{
    text << endl;
    string methodClassName = className + "_" + m->label->messageTarget;

    ///--------------------- Constructor -------------------------------

    text << methodClassName << "::"<< methodClassName <<"(PN_Class *parentObj";

    /// constructor parameters
    list<string>::iterator it_var, it_end;
    it_var = m->label->operands.begin();
    it_end = m->label->operands.end();

    for(; it_var != it_end; it_var++)
    {
        text << ", Token *" << *it_var << "_Operand";
    }

    text << ")" << endl;
    text << "\t:PN_Method_Net(parentObj)";
    text << "{\n";

    if(m->methodNet != NULL)
    {
        text << "\t// ------------------places init----------------\n";

        list<pnPlace*>::iterator it_var, it_end;
        it_var = m->methodNet->places.begin();
        it_end = m->methodNet->places.end();

        for(; it_var != it_end; it_var++)
        {
            GeneratePlaceInitCpp(text, *it_var);
        }

    /// method arguments initialization
        list<string>::iterator it_var_arg, it_end_arg;
        it_var_arg = m->label->operands.begin();
        it_end_arg = m->label->operands.end();

        for(; it_var_arg != it_end_arg; it_var_arg++)
        {
            text << "\tlist<Token *> " << *it_var_arg << "_init;\n";
            text << "\t" << *it_var_arg << "_init.push_back(" << *it_var_arg << "_Operand);\n";
            text << "\t" << *it_var_arg << "->PushTokens(&" << *it_var_arg << "_init);\n\n";
        }


        text << "\t// ----------------transitions init---------------\n";

        list<pnTransition*>::iterator it_var2, it_end2;
        it_var2 = m->methodNet->transitions.begin();
        it_end2 = m->methodNet->transitions.end();

        for(; it_var2 != it_end2; it_var2++)
        {
            GenerateTransitionInitCpp(text, *it_var2);
        }

    }
    text << "}\n\n";


    ///---------------------- Destructor--------------------------------

    text << methodClassName <<"::~"<< methodClassName <<"()\n";
    text << "{\n";

    if(m->methodNet != NULL)                            /// deallocate places
    {
        list<pnPlace*>::iterator it_var, it_end;
        it_var = m->methodNet->places.begin();
        it_end = m->methodNet->places.end();

        for(; it_var != it_end; it_var++)
        {
            if ((*it_var)->name == "return")
            {
                text << "\tdelete _return;\n";    ///place free (special return case)
            }
            else
            {
                text << "\tdelete " << (*it_var)->name << ";\n";    ///place free
            }

        }
    }
    text << "}\n\n";
}


/** @brief GenerateNewInstance
  *
  * @todo: document this function
  */
string OperandVarCpp::GenerateNewInstance()
{
    ostringstream text;
    if (isupper(this->varName.at(0)))
    {
        text << "new PN_Operand_Token("<< this->varName << "::GetInstantiator())";
    }
    else if(this->varName == "self")
    {
        text << "new PN_Operand_Token(new PNToken(this->GetSelf()))";
    }
    else
    {
        text << "new PN_Operand_Var(\""<< this->varName << "\")";
    }
    return text.str();
}

/** @brief GenerateNewInstance
  *
  * @todo: document this function
  */
string OperandTokenCpp::GenerateNewInstance()
{
    ostringstream text;
    text << "new PN_Operand_Token("<< this->lit->GenerateNewInstance() <<")";
    return text.str();
}







