#ifndef CPPGENERATOR_H
#define CPPGENERATOR_H

#include "PNparse.h"

class pnClassCpp : public pnClass
{

protected:
    virtual void addMethod(pnClassMethod* net);
    virtual void addConstructor(pnClassMethod* net);
    virtual void addSyncPort(pnClassSync* sync);
public:
    list<pnClassMethod *> constructors;
    list<pnClassMethod *> methods;
    list<pnClassSync *> syncPorts;
    pnClassCpp(ClassHead *head, pnNet *_objectNet, pnClassContent *content) :pnClass(head, _objectNet){ this->InitContent(content);}
    virtual ~pnClassCpp();
    list<string> GetRequiredClasses();
    virtual void Print();
};

void GeneratePlaceInitCpp(ostringstream& text, pnPlace* places);
void GenerateTransitionInitCpp(ostringstream& text, pnTransition* trans);
void GenerateSyncPortInitCpp(ostringstream& text, pnClassSync *port);
void GenerateConditionInitCpp(ostringstream& text, pnCondition* trans, string listName);
string GenerateMultisetCpp(ostringstream& text, pnMultiset* set, string varPrefix);

void GenerateMethodHeaderCpp(ostringstream& text, pnClassMethod *m, string className);
void GenerateMethodContentCpp(ostringstream& text, pnClassMethod *m, string className);

void GenerateSendMessage(ostringstream& text, pnClassCpp *c);
void GenerateSimulateMessage(ostringstream& text, pnClassCpp *c);

class CppGenerator : public PN_AST_Base
{
public:
    CppGenerator();
    ~CppGenerator();
    void GenerateCode(string dir, string programName);

    virtual pnClass *CreateClass(ClassHead *head, pnNet *_objectNet, pnClassContent *content);
    virtual void AddClass(ClassHead *head, pnNet *_objectNet, pnClassContent *content);

    virtual pnSetItem *CreateMultisetItemVar(string varName);
    virtual pnSetItem *CreateMultisetItemLiteral(Literal* literal);
    virtual pnMultisetItemList *CreateMultisetItemList();
    virtual IntegerLiteral *CreateMultisetItemInt(int value);
    virtual CharLiteral *CreateMultisetItemChar(char value);
    virtual StringLiteral *CreateMultisetItemString(string value);
    virtual SymbolLiteral *CreateMultisetItemSymbol(string value);

    virtual pnMessageUnary *CreateUnaryMessage(string msg, pnMessageUnary *nextMsg);
    virtual pnMessageBinary *CreateBinaryMessage(string msg, pnOperand* op, pnMessageUnary *opChain);
    virtual pnMessageKeys *CreateKeysMessage();

    virtual pnOperandVar *CreateOperandVar(string varName);
    virtual pnOperandLiteral *CreateOperandToken(Literal *token);

protected:
    virtual list<pnClass*> GetClasses()
    {
        list<pnClass*>result;
        list<pnClassCpp*>::iterator it_var, it_end;
        it_var = classes.begin();
        it_end = classes.end();

        for(; it_var != it_end; it_var++)
        {
            result.push_back(*it_var);
        }
        return result;
    }
private:
    list<pnClassCpp *> classes;
    void GeneratePNHeader(pnClassCpp *c, string dir);
    void GeneratePNClass(pnClassCpp *c, string dir);
    void GenerateMain(string dir);
    void GenerateMake(string dir, string programName);
};

///-----------------------------primitive generators------------------------------------------------


    ///-------------------- Multiset ----------------------
class pnMultisetItemVarCpp : public pnMultisetItemVar
{
public:
    pnMultisetItemVarCpp(string value) : pnMultisetItemVar(value) {}
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level);
};

class pnMultisetItemLiteralCpp : public pnMultisetItemLiteral
{
public:
    pnMultisetItemLiteralCpp(Literal *value): pnMultisetItemLiteral(value) {}
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level );
};

class pnMultisetItemListCpp : public pnMultisetItemList
{
public:
    pnMultisetItemListCpp() : pnMultisetItemList() {}
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level);
};

    ///-----------------------------Literals-------------------------------
class IntegerLiteralCpp : public IntegerLiteral
{
public:
    IntegerLiteralCpp(int value) : IntegerLiteral(value) {}
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level );
    virtual string GenerateNewInstance();
};

class CharLiteralCpp : public CharLiteral
{
public:
    CharLiteralCpp(char value) : CharLiteral(value) {}
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level );
    virtual string GenerateNewInstance();
};

class StringLiteralCpp : public StringLiteral
{
public:
    StringLiteralCpp(string value) : StringLiteral(value) {}
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level );
    virtual string GenerateNewInstance();
};

class SymbolLiteralCpp : public SymbolLiteral
{
public:
    SymbolLiteralCpp(string value) : SymbolLiteral(value) {}
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level );
    virtual string GenerateNewInstance();
};

class BoolLiteralCpp : public BoolLiteral
{
    string ToString();
public:
    BoolLiteralCpp(bool value) : BoolLiteral(value) {}
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level );
    virtual string GenerateNewInstance();

};

class NilLiteralCpp : public NilLiteral
{
public:
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level );
    virtual string GenerateNewInstance();
};

    ///-----------------------Messages-------------------------------------
class UnaryMessageCpp : public pnMessageUnary
{
public:
    UnaryMessageCpp(string msg, pnMessageUnary *nextMsg) : pnMessageUnary(msg, nextMsg) {}
    virtual void GenerateSetAction(ostringstream& text, string actionName, pnOperand *target);
};

class BinaryMessageCpp : public pnMessageBinary
{
public:
    BinaryMessageCpp(string msg, pnOperand *op, pnMessageUnary *opChain) : pnMessageBinary(msg, op, opChain) {}
    virtual void GenerateSetAction(ostringstream& text, string actionName, pnOperand *target);
};

class KeysMessageCpp : public pnMessageKeys
{
public:
    KeysMessageCpp() : pnMessageKeys() {}
    virtual void GenerateSetAction(ostringstream& text, string actionName, pnOperand *target);
};

    ///----------------------Operands-------------------------------
class OperandTokenCpp : public pnOperandLiteral
{
public:
    OperandTokenCpp(Literal *literal) : pnOperandLiteral(literal) {}
    string GenerateNewInstance();
};

class OperandVarCpp : public pnOperandVar
{
public:
    OperandVarCpp(string varName) : pnOperandVar(varName) {}
    string GenerateNewInstance();
};

#endif // CPPGENERATOR_H
