#ifndef GENERATORFACTORY_H
#define GENERATORFACTORY_H

#include <list>
#include <string>

class pnMessageUnary;
class pnMessageBinary;
class pnMessageKeys;
class pnOperand;

class pnOperandVar;
class pnOperandLiteral;

class pnSetItem;
class pnMultisetItemList;

class Literal;
class IntegerLiteral;
class CharLiteral;
class StringLiteral;
class SymbolLiteral;

class pnClass;
class ClassHead;
class pnNet;

struct pnClassContent;

using namespace std;

class PN_AST
{
public:

    virtual void Print()=0;

    virtual void GenerateCode(string dir, string programName)=0;

    virtual pnClass *CreateClass(ClassHead *head, pnNet *_objectNet, pnClassContent *content)=0;
    virtual void SetStartClass(string className)=0;
    virtual void AddClass(ClassHead *head, pnNet *_objectNet, pnClassContent *content)=0;

    virtual pnSetItem *CreateMultisetItemVar(string varName)=0;
    virtual pnSetItem *CreateMultisetItemLiteral(Literal* literal)=0;
    virtual pnMultisetItemList *CreateMultisetItemList()=0;

    virtual IntegerLiteral *CreateMultisetItemInt(int value)=0;
    virtual CharLiteral *CreateMultisetItemChar(char value)=0;
    virtual StringLiteral *CreateMultisetItemString(string value)=0;
    virtual SymbolLiteral *CreateMultisetItemSymbol(string value)=0;

    virtual pnMessageUnary *CreateUnaryMessage(string msg, pnMessageUnary *nextMsg)=0;
    virtual pnMessageBinary *CreateBinaryMessage(string msg, pnOperand* op, pnMessageUnary *opChain)=0;
    virtual pnMessageKeys *CreateKeysMessage()=0;

    virtual pnOperandVar *CreateOperandVar(string varName) = 0;
    virtual pnOperandLiteral *CreateOperandToken(Literal *literal) = 0;

    virtual bool DoSemantics()= 0;

};

/** Singleton generator
    *
    */
class GeneratorImplementation
{
public:
    static GeneratorImplementation* Instance();

    void SetImplementation(PN_AST *impl) {implementation = impl;}

    void Print();
    void GenerateCode(string dir, string programName);

    pnClass *CreateClass(ClassHead *head, pnNet *_objectNet, pnClassContent *content);
    void SetStartClass(string className);
    void AddClass(ClassHead *head, pnNet *_objectNet, pnClassContent *content);

    pnSetItem *CreateMultisetItemVar(string varName);
    pnSetItem *CreateMultisetItemLiteral(Literal* literal);
    pnMultisetItemList *CreateMultisetItemList();

    IntegerLiteral *CreateMultisetItemInt(int value);
    CharLiteral *CreateMultisetItemChar(char value);
    StringLiteral *CreateMultisetItemString(string value);
    SymbolLiteral *CreateMultisetItemSymbol(string value);

    pnMessageUnary *CreateUnaryMessage(string msg, pnMessageUnary *nextMsg);
    pnMessageBinary *CreateBinaryMessage(string msg, pnOperand* op, pnMessageUnary *opChain);
    pnMessageKeys *CreateKeysMessage();

    pnOperandVar *CreateOperandVar(string varName);
    pnOperandLiteral *CreateOperandToken(Literal *literal);

    bool DoSemantics();

private:
    PN_AST *implementation;
    GeneratorImplementation() {implementation = NULL;}
    GeneratorImplementation(GeneratorImplementation const&);
    void operator=(GeneratorImplementation const&);
    static GeneratorImplementation* m_Instance;

};

#endif // GENERATORFACTORY_H
