#include "Bindings.h"
#include "Token.h"

#include <typeinfo>
#include <vector>
#include <iostream>


/** @brief BindingFinder
  *
  * @todo: document this function
  */
BindingFinder::BindingFinder(PN_Condition *_cond)
{
    this->cond = _cond;

    BindingState * initState = new BindingState(cond->targetPlace, cond->set);
    this->states.push_back(initState);

    this->partialBinding = new PN_TransitionBinding();
}

/** @brief Print
  *
  * @todo: document this function
  */
void PN_TransitionBinding::Print()
{
    if(this->bindings.empty())
    {
        cout << "no variables in valid binding\n";
    }
    else
    {
        list<PN_Binding>::iterator it_var2, it_end2;
        it_var2 = this->bindings.begin();
        it_end2 = this->bindings.end();
        while(it_var2 != it_end2)
        {
            cout << (*it_var2).varName << " => ";
            (*it_var2).target->Print();
            cout << endl;
            it_var2++;
        }
    }
}

/** @brief Merge
  *
  * Merge this binding with another,
  *
  * @return false if conflict occurs, true otherwise
  *
  */
bool PN_TransitionBinding::Merge(PN_TransitionBinding* other)
{
    list<PN_Binding> addList;

    list<PN_Binding>::iterator it_var, it_end;
    it_var = other->bindings.begin();
    it_end = other->bindings.end();
    while(it_var != it_end)
    {
        bool varFound = false;
        list<PN_Binding>::iterator it_var2, it_end2;
        it_var2 = this->bindings.begin();
        it_end2 = this->bindings.end();
        while(it_var2 != it_end2)
        {
            if(it_var->varName == it_var2->varName)
            {
                if(compare_tokens_cmp(it_var->target, it_var2->target) != 0)
                {
                    return false;
                }
                else
                {
                    varFound = true;
                    it_var2 = it_end2;
                }
            }
            else
            {
                it_var2++;
            }
        }
        if(!varFound)
        {
            addList.push_back(*it_var);
        }
        it_var++;
    }

    this->bindings.insert(this->bindings.end(), addList.begin(), addList.end());
    return true;
}

list<PN_TransitionBinding*> *BindingFinder::GetBindingsNew()
{
    list<PN_TransitionBinding*> *result = new list<PN_TransitionBinding*>();
    bool noMoreBingings = false;
    do
    {
        PN_TransitionBinding* next = this->GetNextBinding();
        if(next != NULL)
        {
//            cout << "next binding :";
//            next->Print();
//            cout << endl;
            result->push_back(next);
        }
        else
        {
            noMoreBingings = true;
        }
    }
    while (!noMoreBingings);
    return result;
}

/** @brief Tries to add next Binding to Binding collection
 *
 * @param partialBinding Already constructed partial binding
 * @param next next binding to add
 * @return true if next binding was added, false if binding conflict occurs
 *
 * Tries to add next Binding to Binding collection. Fails if there already is binding for variable but to different value.
 * If Binding collection contains binding to same variable and same value, does nothing, but returns true.
 */
bool TryAddNextBinding(list<PN_Binding> *partialBinding, PN_Binding next)
{
    list<PN_Binding>::iterator it_var, it_end;

    it_var = partialBinding->begin();
    it_end = partialBinding->end();
    while(it_var != it_end)
    {
        if(next.varName == it_var->varName)
        {
            return compare_tokens_cmp(next.target, it_var->target) == 0;
        }
        it_var++;
    }
    partialBinding->push_back(next);
    return true;
}

list<PN_Binding> GetListBinding(ListToken *l, PN_MultiItemList *m, list<PN_Binding> partialBinding)
{
    if(l->collection.size() == m->listCol.size())
    {
        list<Token*>::iterator it_var_token;
        it_var_token = l->collection.begin();
        list<PN_MultiItem*>::iterator it_var_mItem;
        it_var_mItem = m->listCol.begin();

        while (it_var_token != l->collection.end())
        {
            if ((*it_var_mItem)->GetType() != variable && (*it_var_mItem)->GetType() != mlist )
            {
                Token* mItemToken = (*it_var_mItem)->CreateTokenPrototype(partialBinding);
                if (compare_tokens_cmp(mItemToken, *it_var_token ) != 0)
                {
                    delete mItemToken;
                    throw PN_ListNotBindable();
                }
                else delete mItemToken;
            }
            else if ((*it_var_mItem)->GetType() == variable)
            {
                string varName = ((PN_MultiItemVariable*)(*it_var_mItem))->varname;
                if(!TryAddNextBinding(&partialBinding, PN_Binding(varName, *it_var_token)))
                {
                    throw PN_ListNotBindable();
                }
            }
            else
            {
                if((*it_var_token)->GetType() == list_token)
                {
                    list<PN_Binding> listBinding = GetListBinding((ListToken*)*it_var_token,
                                                   (PN_MultiItemList*)*it_var_mItem,
                                                   partialBinding);
                    partialBinding = listBinding;
                }
                else
                {
                    throw PN_ListNotBindable();
                }
            }
            it_var_token++;
            it_var_mItem++;
        }
    }
    return partialBinding;
}

/** @brief GetNextBinding
  *
  * Tries to find next full binding of associated PN_Condition, if variable is encounted,
  * genetates new states to internal storage
  *
  * @return: new binding or NULL, if no binding could be found
  */
PN_TransitionBinding* BindingFinder::GetNextBinding()
{
    if(this->states.empty())
    {
        return NULL; // nothing left to bind
    }

    BindingState *act_state = states.front(); // get next partially processed state
    states.pop_front();

    while (!act_state->multiSetToBind.empty())
    {
        PN_MultiItem *item = act_state->multiSetToBind.front(); // process next Multiset Item form Multiset
        act_state->multiSetToBind.pop_front();

        if( item->GetType() != variable && item->GetType() != mlist ) /// if Literal, tries to find literal
        {
            PN_MultiItemPrimitive *primitive = (PN_MultiItemPrimitive *)item;
            list<Token*> removeList;
            try
            {
                try
                {
                    removeList = primitive->Tokenize(act_state->foundBindings);
                    act_state->GetPlaceCopy()->PopTokens(&removeList);
                    DisposeRefList<Token*>(removeList);
                }
                catch (PN_CantResolve_Exception *ex)
                {
                    /// quantity variable is still unbinded
                    Token *itemToDelete = primitive->CreateTokenPrototype();
                    int itemCount = act_state->GetPlaceCopy()->PopTokenGreedy(itemToDelete);
                    delete itemToDelete;
                    /// create new var binding with greedy pop quantity
                    Token *token = new IntToken(itemCount);
                    act_state->foundBindings.push_back(PN_Binding(item->QuantityVarName(), token));
                }
            }
            catch (PN_PlaceNotContainException ex) // if Binding literals fails, skip to next partially processed state
            {
                DisposeRefList<Token*>(removeList);
                delete act_state;
                if(states.empty())
                    return NULL;
                act_state = states.front();
                states.pop_front();
            }
        }
        else if (item->GetType() == variable)      /// if Variable, generate all new states, that could come from possible binding
        {
            list<BindingState*> statesToAdd;
            string var = ((PN_MultiItemVariable*)item)->varname;
            int target_q = item->GetQuantity();
            if (item->IsQuantityRef())
            {
                Token *bindValue = FindBinding(&(act_state->foundBindings), item->QuantityVarName());
                IntToken *intRefQuantity = dynamic_cast<IntToken*>(bindValue);
                if (intRefQuantity != NULL)
                {
                    target_q = intRefQuantity->value;
                }
            }
            int act_q = 0;

            AllTokensIterator *it = act_state->GetPlaceCopy()->GetIter();
            Token *act_token = NULL;

//            cout<< "next variable (" << target_q << " ` " <<  var << "), actual state:\n";
//            act_state->Print();

            bool greedyFinished = false;
            while(!it->IsAtEnd()) /// we must find all possible binding of variable
            {
                Token* next_token = **it;
                if (compare_tokens_cmp(act_token, next_token) == 0) // Token in place is same as predecessor
                {
                    act_q++; // increment counter
                    it->Increment();
                }
                else // token is different => set new token
                {
                    // search is greedy, mark it finished (reset counter later)
                    if(target_q == -1)
                    {
                        greedyFinished = true;
                    }
                    // if not greedy, reset counter
                    else
                    {
                        act_q = 0;
                    }
                    act_token = next_token;
                }
                if (-1 == target_q && it->IsAtEnd())
                {
                    greedyFinished = true;
                }
                // if var could be bind (enough Tokens are at PN place), create new partially processed state according to values
                if(act_q == target_q || greedyFinished)
                {
                    // make copy of actual binding, because we will bind up to 2 variables (must avoid half-effect)
                    list<PN_Binding> partB(act_state->foundBindings);

                    // first try add variable value
                    if(TryAddNextBinding(&partB, PN_Binding(var, act_token)))
                    {
                        // logical implication greedyFinished => quantity can be bind (if not greedy, don't care)
                        if (!greedyFinished || TryAddNextBinding(&partB, PN_Binding(item->QuantityVarName(), new IntToken(act_q))))
                        {
                            // binding was added or merged -> create new state and simulate binding upon copy place
                            BindingState* newState = new BindingState(*act_state);
                            newState->foundBindings = partB;

                            list<Token*> removeList;
                            for (int i = 0; i< act_q; i++)
                            {
                                removeList.push_back(act_token->Copy());
                            }
                            newState->GetPlaceCopy()->PopTokens(&removeList);
                            statesToAdd.push_back(newState);
                        }
                    }
                    if (greedyFinished)
                    {
                        act_q = 0;
                        greedyFinished = false;
                    }
                }
            }
            states.insert(states.begin(), statesToAdd.begin(), statesToAdd.end());
            /// get next state - if no binding was found, abandons this branch, otherwise pops state with first variable binding
            delete it;
            delete act_state;
            if(states.empty())
                return NULL;
            act_state = states.front();
            states.pop_front();
        }
        else if (item->GetType() == mlist)
        {
//            list<Token*> removeList = item->Tokenize();
//            if(!removeList.empty())                         /// list contains only Literals (fast version)
//            {
//                try
//                {
//                    act_state->GetPlaceCopy()->PopTokens(&removeList, false);
//                }
//                catch (PN_PlaceNotContainException ex) // if Binding literals fails, skip to next partially processed state
//                {
//                    DisposeRefList<Token*>(removeList);
//                    delete act_state;
//                    if(states.empty())
//                        return NULL;
//                    act_state = states.front();
//                    states.pop_front();
//                }
//            }
//            else                                            /// MultiItemList contains variables
            {
                list<BindingState*> statesToAdd;
                PN_MultiItemList *act_list = ((PN_MultiItemList*)item);

                int target_q = item->GetQuantity();
                if (item->IsQuantityRef())
                {
                    Token *bindValue = FindBinding(&(act_state->foundBindings), item->QuantityVarName());
                    IntToken *intRefQuantity = dynamic_cast<IntToken*>(bindValue);
                    if (intRefQuantity != NULL)
                    {
                        target_q = intRefQuantity->value;
                    }
                }

                list<ListToken*> placeLists = act_state->GetPlaceCopy()->listCol;
                list<ListToken*>::iterator it_var, it_end;
                it_var = placeLists.begin();
                it_end = placeLists.end();
                int act_q = 0;
                ListToken *act_token = NULL;
                bool greedyFinished = false;
                while(it_var != it_end) // we must try all lists form place
                {
                    if (compare_tokens_cmp(act_token, *it_var) == 0) // Token in place is same as predecessor
                    {
                        act_q++; // increment quantity counter
                        it_var++;
                    }
                    else // token is different => set new token, reset quantity counter
                    {
                        // search is greedy, mark it finished (reset counter later)
                        if(target_q == -1)
                        {
                            greedyFinished = true;
                        }
                        // if not greedy, reset counter
                        else
                        {
                            act_q = 0;
                        }
                        act_token = *it_var;
                    }
                    // if vars found in list could be bind, create new partially processed state according to values
                    if(act_q == target_q || greedyFinished)
                    {
                        list<PN_Binding> listBinding;
                        bool listBinded;

                        try
                        {
                            listBinding = GetListBinding(act_token, act_list, act_state->foundBindings);
                            listBinded = true;
                        }
                        catch(PN_ListNotBindable ex)
                        {
                            listBinded = false;
                        }

                        if(listBinded)
                        {
                             // logical implication greedyFinished => quantity can be bind (if not greedy, don't care)
                            if (!greedyFinished || TryAddNextBinding(&listBinding, PN_Binding(item->QuantityVarName(), new IntToken(act_q))))
                            {
                                BindingState* newState = new BindingState(*act_state);
                                newState->foundBindings = listBinding;
                                list<Token*> removeList2;
                                for (int i = 0; i< act_q; i++)
                                {
                                    removeList2.push_back(act_token->Copy());
                                }
                                newState->GetPlaceCopy()->PopTokens(&removeList2);
                                statesToAdd.push_back(newState);
                            }
                        }
                        if (greedyFinished)
                        {
                            act_q = 0;
                            greedyFinished = false;
                        }
                    }
                }
                states.insert(states.begin(), statesToAdd.begin(), statesToAdd.end());
                /// get next state - if no binding was found, abandons this branch, otherwise pops state with first var binding
                delete act_state;
                if(states.empty())
                    return NULL;
                act_state = states.front();
                states.pop_front();

            } // end mList with variables
        } // end mList case
    }
    /// succesful binding at this point (processed all items from multiset)
    PN_TransitionBinding *result = new PN_TransitionBinding();
    result->bindings = act_state->foundBindings;
    return result;
}


/** @brief Print
  *
  * @todo: document this function
  */
void BindingState::Print()
{
    this->placeCopy.PrintContent();
    cout << "bindings: ";
    list<PN_Binding>::iterator it_var, it_end;
    it_var = this->foundBindings.begin();
    it_end = this->foundBindings.end();
    while(it_var != it_end)
    {
        cout << it_var-> varName << " = ";
        it_var->target->Print();
        cout << ", ";
        it_var++;
    }
    cout << endl;

    cout << "items to bind: ";
    list<PN_MultiItem*>::iterator it_var2, it_end2;
    it_var2 = this->multiSetToBind.begin();
    it_end2 = this->multiSetToBind.end();
    while(it_var2 != it_end2)
    {
        (*it_var2)->Print();
        cout << ", ";
        it_var2++;
    }
    cout << endl;
}

void PrintBindings(list<PN_Binding>& binds)
{
    list<PN_Binding>::iterator it_var, it_end;
    it_var = binds.begin();
    it_end = binds.end();
    cout << "binds: ";
    while(it_var != it_end)
    {
        cout << it_var-> varName << " = ";
        it_var->target->Print();
        cout << ", ";
        it_var++;
    }
    cout << endl;
}

void DisposeBindings(list<PN_Binding>& binds)
{
    list<PN_Binding>::iterator it_var, it_end;
    it_var = binds.begin();
    it_end = binds.end();
    while(it_var != it_end)
    {
        Token *del = it_var->target;
        delete del;
        it_var++;
    }
}

/** @brief SetPartialBinding
  *
  * @todo: document this function
  */
void BindingFinder::SetPartialBinding(PN_TransitionBinding* b)
{
    delete this->partialBinding;
    if (b != NULL)
    {
        this->partialBinding = b;
    }
    else
    {
        this->partialBinding = new PN_TransitionBinding();
    }


    DisposeRefList(this->states);
    this->states.clear();

    BindingState * initState = new BindingState(cond->targetPlace, cond->set);
    initState->foundBindings = this->partialBinding->bindings;
    this->states.push_back(initState);
}






