#ifndef PN_MULTISET_H
#define PN_MULTISET_H

#include <list>
#include <string>
#include "Token.h"

using namespace std;

class PN_MultiItem;

typedef enum {integer, str, character, symbol, variable, mlist} PN_MultiItemType;

class PN_Multiset
{
    //friend class BindingFinder
    public:
        PN_Multiset();
        virtual ~PN_Multiset();
        void AddItem(PN_MultiItem *item);
        list<PN_MultiItem*> items;
    protected:
    private:

};

class PN_MultiItem
{
protected:
    int quantity;
    bool quantityIsRef;
    string refVarName;
public:
    PN_MultiItem (int _quantity) {quantity = _quantity; quantityIsRef = false;}
    PN_MultiItem (string refVar) {refVarName = refVar; quantityIsRef = true; quantity = -1;}
    virtual ~PN_MultiItem() {}
    int GetQuantity() {return this->quantity;}
    bool IsQuantityRef () {return this->quantityIsRef;}
    string QuantityVarName() {return this->refVarName;}

    virtual PN_MultiItemType GetType()=0;

    list<Token*> Tokenize(list<PN_Binding>& binding)
    {
        int effectiveQuantity = quantity;
        if(quantityIsRef)
        {
            Token *bindValue = FindBinding(&binding, refVarName);
            IntToken *intRefQuantity = dynamic_cast<IntToken*>(bindValue);
            if (intRefQuantity == NULL)
            {
                throw new PN_CantResolve_Exception(refVarName);
            }
            effectiveQuantity = intRefQuantity->value;
        }

        list<Token*> result;
        Token *prototype = this->CreateTokenPrototype(binding);
        result.push_back(prototype);
        for(int i = 1; i < effectiveQuantity; i++)
        {
            result.push_back(prototype->Copy());
        }
        return result;
    }
    virtual Token* CreateTokenPrototype(list<PN_Binding>& binding)=0;
    virtual void Print() = 0;
};

class PN_MultiItemPrimitive : public PN_MultiItem
{
public:
    PN_MultiItemPrimitive (int _quantity) :PN_MultiItem(_quantity) {}
    PN_MultiItemPrimitive (string _quantityRef) :PN_MultiItem(_quantityRef) {}
    virtual Token* CreateTokenPrototype()=0;
    virtual Token* CreateTokenPrototype(list<PN_Binding>& binding)
    {
        return this->CreateTokenPrototype();
    }
    using PN_MultiItem::Tokenize;
    list<Token*> Tokenize()
    {
        list<Token*> result;
        Token *prototype = this->CreateTokenPrototype();
        result.push_back(prototype);
        for(int i = 1; i < this->quantity; i++)
        {
            result.push_back(prototype->Copy());
        }
        return result;
    }
};

class PN_MultiItemInt : public PN_MultiItemPrimitive
{
    public:
        int value;
        PN_MultiItemInt (int _quantity, int _value) : PN_MultiItemPrimitive(_quantity) {value = _value;}
        PN_MultiItemInt (string _quantityRef, int _value) : PN_MultiItemPrimitive(_quantityRef) {value = _value;}
        PN_MultiItemType GetType() {return integer;}
        virtual Token* CreateTokenPrototype();
        virtual void Print() {cout << quantity << "`" << value;}
};

class PN_MultiItemString : public PN_MultiItemPrimitive
{
    public:
        string value;
        PN_MultiItemString (int _quantity, string _value) : PN_MultiItemPrimitive(_quantity) {value = _value;}
        PN_MultiItemString (string _quantityRef, string _value) : PN_MultiItemPrimitive(_quantityRef) {value = _value;}
        PN_MultiItemType GetType() {return str;}
        virtual Token* CreateTokenPrototype();
        virtual void Print() {cout << quantity << "`'" << value << "'";}
};

class PN_MultiItemChar : public PN_MultiItemPrimitive
{
    public:
        char value;
        PN_MultiItemChar (int _quantity, char _value) : PN_MultiItemPrimitive(_quantity) {value = _value;}
        PN_MultiItemChar (string _quantityRef, char _value) : PN_MultiItemPrimitive(_quantityRef) {value = _value;}
        PN_MultiItemType GetType() {return character;}
       virtual Token* CreateTokenPrototype();
        virtual void Print() {cout << quantity << "`$" << value;}
};

class PN_MultiItemSymbol : public PN_MultiItemPrimitive
{
    public:
        string value;
        PN_MultiItemSymbol (int _quantity, string _value) : PN_MultiItemPrimitive(_quantity) {value = _value;}
        PN_MultiItemSymbol (string _quantityRef, string _value) : PN_MultiItemPrimitive(_quantityRef) {value = _value;}
        PN_MultiItemType GetType() {return symbol;}
        virtual Token* CreateTokenPrototype();
        virtual void Print() {cout << quantity << "`#" << value;}
};

class PN_MultiItemVariable : public PN_MultiItem
{
    public:
        string varname;
        PN_MultiItemVariable (int _quantity, string _value) : PN_MultiItem(_quantity) {varname = _value;}
        PN_MultiItemVariable(string _quantityRef, string _value) : PN_MultiItem(_quantityRef) {varname = _value;}
        PN_MultiItemType GetType() { return variable;}
        virtual Token* CreateTokenPrototype(list<PN_Binding>& binding);
        virtual void Print() {cout << quantity << "`" << varname;}
};

class PN_MultiItemList : public PN_MultiItem
{
    public:
        list<PN_MultiItem*> listCol;
        PN_MultiItemList (int _quantity) : PN_MultiItem(_quantity) {}
        PN_MultiItemList(string _quantityRef) : PN_MultiItem(_quantityRef) {}
        virtual ~PN_MultiItemList () { DisposeRefList(this->listCol);}
        void AddItem(PN_MultiItem *item) {listCol.push_back(item);}
        PN_MultiItemType GetType() {return mlist;}
        virtual Token* CreateTokenPrototype(list<PN_Binding>& binding);
        virtual void Print() ;
};

#endif // PN_MULTISET_H
