#ifndef PN_PLACE_H
#define PN_PLACE_H

#include "PN_Multiset.h"
#include <string>

class AllTokensIterator;

class PN_PlaceNotContainException : PN_Error_Exception
{
public:
    PN_PlaceNotContainException(string msg) : PN_Error_Exception(msg) {}
};

class PN_Place
{
    friend class AllTokensIterator;
    friend class BindingFinder;
    public:
        PN_Place(std::string _name): name(_name){}
        PN_Place(PN_Multiset *s, std::string _name);
        virtual ~PN_Place();
        void Clear();
        void PrintContent();
        string GetName() {return name;}
        void PopTokens(list<Token*> *tokens);
        virtual void PushTokens(list<Token*> *tokens);
        int PopTokenGreedy(Token *token);
        AllTokensIterator *GetIter();
    protected:
    private:
        std::string name;
        list<IntToken *> integersCol;
        list<StringToken *> stringCol;
        list<CharToken *> charCol;
        list<SymbolToken *> symbolCol;
        list<Token *> refCol;
        list<ListToken *> listCol;
};

class AllTokensIterator
{
    enum TokenType {integers, chars, strings, symbols, refs, lists, end};
    private:
        TokenType act_type;
        list<IntToken*>::iterator act_int;
        list<StringToken*>::iterator act_string;
        list<CharToken*>::iterator act_char;
        list<SymbolToken*>::iterator act_symbol;
        list<Token*>::iterator act_ref;
        list<ListToken*>::iterator act_list;
        PN_Place *parent;
    public:
        AllTokensIterator(PN_Place *p);
        void Increment(bool do_not_incrementate = false);
        Token* operator*() const;
        bool IsAtEnd () {return this->act_type == end;}
        void EraseActual();
};





#endif // PN_PLACE_H
