#include "PN_Smalltalk.h"

#include "Token.h"


Token* PN_MetaClass_Time::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token *result = NULL;

    ///-------------------------Unary messages--------------------------

    if(message == "millisecondClockValue")
    {
        result = new TimeToken(new Time(clock()));
    }
    return result;

}

Token* TimeToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{

    if( message == "-")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        arg1 = args.front()->Resolve(binds);
        TimeToken* secondTime = dynamic_cast<TimeToken*>(arg1);
        if(secondTime != NULL)
        {
            return new TimeToken(new Time( this->value->m_time - secondTime->value->m_time));
        }

    }
    return Token::SendMessage(message, args, binds);
}

Token *PN_MetaClass_Transcript::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    if(message == "cr")
    {
        cout << "\r";
        return new NilToken();
    }
    if(message == "lf")
    {
        cout << "\n";
        return new NilToken();
    }
    if( message == "show:")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        arg1 = args.front()->Resolve(binds);
        cout << arg1->AsText();
        return new NilToken();
    }
    return NULL;
}

Token *PN_MetaClass_OrderedCollection::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    if(message == "new")
    {
        return new OrderedCollectionToken(new OrderedCollection());
    }
    else
    {
        throw new PN_MessageNotUnderstood_Exception (this->AsText(), message);
    }
}

Token *OrderedCollection::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    if(message == "removeFirst")
    {
        if (this->m_collection.empty())
        {
            throw string("Failed removing from collection");
        }
        else
        {
            this->m_collection.erase(m_collection.begin());
            return new NilToken();
        }
    }
    if(message == "removeLast")
    {
        if (this->m_collection.empty())
        {
            throw string("Failed removing from collection");
        }
        else
        {
            this->m_collection.pop_back();
            return new NilToken();
        }
    }

    if(message == "at:")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        arg1 = args.front()->Resolve(binds);
        IntToken* index = dynamic_cast<IntToken*>(arg1);
        if(index != NULL)
        {
            return (*(m_collection.begin() + index->value))->Copy();
        }
    }
    if(message == "at:put:")
    {
        if (args.size() < 2)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        arg1 = args.front()->Resolve(binds);
        IntToken* index = dynamic_cast<IntToken*>(arg1);
        if(index != NULL)
        {
            Token * anObject = (*(args.begin()++))->Resolve(binds);
            m_collection[index->value] = anObject->Copy();
            return anObject->Copy();
        }
    }
    if(message == "removeAt:")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        arg1 = args.front()->Resolve(binds);
        IntToken* index = dynamic_cast<IntToken*>(arg1);
        if(index != NULL)
        {
            m_collection.erase(m_collection.begin() + index->value);
            return new NilToken();
        }
    }
    if(message == "add:" || message == "addLast:")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        arg1 = args.front()->Resolve(binds);
        this->m_collection.push_back(arg1->Copy());
        return arg1->Copy();
    }
    if(message == "addFirst:")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        arg1 = args.front()->Resolve(binds);
        this->m_collection.insert(m_collection.begin(), arg1->Copy());
        return arg1->Copy();
    }
    if(message == "addAll:")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        arg1 = args.front()->Resolve(binds);
        OrderedCollectionToken* otherColToken = dynamic_cast<OrderedCollectionToken*>(arg1);
        if(otherColToken != NULL)
        {
            int colSize = otherColToken->value->m_collection.size();
            for( int i = 0; i < colSize; i++)
            {
                Token *insToken = otherColToken->value->m_collection[i]->Copy();
                this->m_collection.push_back(insToken);
            }

        }
        return arg1->Copy();
    }
    return NULL;
}
string OrderedCollection::AsText()
{

    vector<Token*>::iterator it_var, it_end;
    it_var = this->m_collection.begin();
    it_end = this->m_collection.end();

    ostringstream text;
    text << "(";
    if(it_var != it_end)
    {
        Token * first = (*it_var);
            text << first->AsText();

        it_var++;
    }

    for (;it_var != it_end; it_var++)
    {
        text << ", "<< (*it_var)->AsText();
    }

    text << ")";
    return text.str();
}

Token *OrderedCollectionToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token *result = this->value->SendMessage(message, args, binds);
    if(result == NULL)
    {
        return ReferenceToken::SendMessage(message, args, binds);
    }
    else
    {
        return result;
    }
}
