#ifndef PN_SMALLTALK_H
#define PN_SMALLTALK_H

#include <ctime>
#include <sstream>
#include <memory>
#include "Token.h"

class PN_MetaClass_Time: public ReferenceToken
{
public:
    PN_MetaClass_Time() {}
    virtual ~PN_MetaClass_Time(){}
    virtual Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    virtual void* GetValue() { return NULL; }
    virtual string AsText() { return "Time meta class"; }
    virtual Token* Copy() {return NULL;}
};

class Time
{
public:
    static PN_MetaClass_Time *GetInstantiator() {return new PN_MetaClass_Time();}
    Time(clock_t time) {m_time = time;}
    clock_t m_time;
};

class TimeToken: public ReferenceToken
{
    private:
        Time *value;
    public:
        TimeToken(Time *time) {value = time;}
        virtual ~TimeToken() {}
        string AsText()
        {
            ostringstream text;
            text << ((float)(value->m_time))/CLOCKS_PER_SEC;
            text << "sec";
            return text.str();
        }
        Token* Copy() { return new TimeToken(value);}
        void* GetValue() {return value;}
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
};

class PN_MetaClass_Transcript: public ReferenceToken
{
public:
    PN_MetaClass_Transcript(){}
    virtual ~PN_MetaClass_Transcript(){}
    virtual Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    virtual void* GetValue() { return NULL; }
    virtual string AsText() { return "Transcript meta class"; }
    virtual Token* Copy() {return NULL;}
};

class Transcript
{
public:
    static PN_MetaClass_Transcript *GetInstantiator() {return new PN_MetaClass_Transcript();}
};

class PN_MetaClass_OrderedCollection: public ReferenceToken
{
public:
    PN_MetaClass_OrderedCollection() {}
    virtual ~PN_MetaClass_OrderedCollection(){}
    virtual Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    virtual void* GetValue() { return NULL; }
    virtual string AsText() { return "Ordered collection meta class"; }
    virtual Token* Copy() {return NULL;}
};

class OrderedCollection
{
private:

public:
    OrderedCollection() {}
    ~OrderedCollection() { DisposeRefList(this->m_collection);}
    static PN_MetaClass_OrderedCollection *GetInstantiator() {return new PN_MetaClass_OrderedCollection();}
    virtual Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    string AsText();
    vector<Token*> m_collection;
};

class OrderedCollectionToken: public ReferenceToken
{
    public:
        OrderedCollectionToken(OrderedCollection *col) {value = col;}
        virtual ~OrderedCollectionToken() {}
        virtual string AsText() {return value->AsText();}
        Token* Copy() { return new OrderedCollectionToken(value);}
        void* GetValue() {return value;}
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        OrderedCollection *value;
};






#endif // PN_SMALLTALK_H
