#include "PN_Transition.h"

#include <sstream>
#include <vector>

PN_Checkable_Port::~PN_Checkable_Port()
{
    delete this->guard;
    DisposeRefList(this->preconds);
    DisposeRefList(this->conds);
    DisposeRefList(this->postconds);
}

PN_Transition::PN_Transition(string _name)
{
    this->name = _name;
    this->died = false;
    this->guard = NULL;
    //ctor
}

PN_Transition::~PN_Transition()
{

    DisposeRefList(this->actions);
    //cout << "destructing transition" << name << endl;
    //dtor
}

void PN_Checkable_Port::SetPreconds(list<PN_Condition *> preC)
{
    this->preconds = preC;
}
void PN_Checkable_Port::SetPostconds(list<PN_Condition *> preC)
{
    this->postconds = preC;
}
void PN_Checkable_Port::SetConds(list<PN_Condition *> preC)
{
    this->conds = preC;
}
void PN_Transition::AddAction (PN_Expression *act)
{
    this->actions.push_back(act);
}
/** @brief SetGuard
  *
  * @todo: document this function
  */
void PN_Checkable_Port::SetGuard(PN_Guard *g)
{
    this->guard = g;
}
void PN_Transition::Die()
{
    died = true;
}
bool PN_Transition::IsDead()
{
    return died;
}

PN_TransitionBinding *PN_Checkable_Port::GetBinding(PN_TransitionBinding *init_binds)
{
    PN_TransitionBinding *result = NULL;

    list<PN_Condition*>::iterator it_var, it_end;
    it_var = this->preconds.begin();
    it_end = this->preconds.end();

/*
    PN_Transition *printTrans = dynamic_cast<PN_Transition*>(this);
    if(printTrans != NULL)
    {
        cout << "searching binding in transition " << printTrans->name << endl;
    }
*/
    vector<BindingFinder*> finders;

    for(; it_var != it_end; it_var++)
    {
        finders.push_back(new BindingFinder(*it_var));
    }

    it_var = this->conds.begin();
    it_end = this->conds.end();
    for(; it_var != it_end; it_var++)
    {
        finders.push_back(new BindingFinder(*it_var));
    }

    int act_tier = 0;
    int final_tier = finders.size();

    if (final_tier != 0)
    {
        finders[0]->SetPartialBinding(init_binds);
    }

    while (act_tier != final_tier && act_tier >= 0)
    {
        PN_TransitionBinding *part_b = finders[act_tier]->GetNextBinding();
        if (part_b != NULL)
        {
            act_tier++;
            if(act_tier == final_tier)
            {
                result = part_b;
                if (this->guard != NULL)
                {
                    if(!this->guard->CheckGuard(result->bindings))
                    {
                        delete result;
                        result = NULL;
                        act_tier--;
                    }
                    /*
                    else
                    {
                        result->Print();
                    }
                    */
                }
            }
            else
            {
                finders[act_tier]->SetPartialBinding(part_b);
            }
        }
        else
        {
            act_tier--;
        }
    }
    DisposeRefList(finders);

//    if(result != NULL)
//    {
//        cout << "next binding :";
//        result->Print();
//        cout << endl;
//    }
    return result;
}

/** @brief ExecuteBinding
  *
  * @todo: document this function
  */
void PN_Transition::ExecuteWithBinding(PN_TransitionBinding *binding)
{
    // pop tokens from preconds
    list<PN_Condition*>::iterator it_var, it_end;
    it_var = this->preconds.begin();
    it_end = this->preconds.end();

    for(; it_var != it_end; it_var++)
    {
        (*it_var)->DoAsPrecond(binding->bindings);
    }

    if(this->guard != NULL)
    {
        if(!this->guard->ExecuteGuard(binding->bindings))
        {
            throw new PN_Error_Exception("Guard checked but couldn't be executed.");
        }
    }

    PN_Action *actionInstance = new PN_Action(this->actions, binding->bindings, this->postconds);
// do action

    delete binding;

    bool actionFinished = false;
    try
    {
        actionFinished = actionInstance->Execute();
        if(actionFinished)
        {
            delete actionInstance;
        }
    }
    catch(PN_JoinException *ex)
    {
        delete actionInstance;
        throw ex;
    }
}

/** @brief Continue
  *
  * @todo: document this function
  */
bool PN_Action::Continue(Token *jMsgResult)
{
//    cout << "action continues with value ";
//    jMsgResult->Print();
//    cout << endl;
    (*act_expr)->UpdateNamespace(this->actionNamespace, jMsgResult);

    act_expr++;
    return this->Execute();
}

/** @brief Execute
  *
  * @todo: document this function
  */
bool PN_Action::Execute()
{
    while(this->act_expr != this->expr.end())
    {
        try
        {
            (*act_expr)->ExecuteAction(this->actionNamespace, false);
            act_expr++;
        }
        catch (PN_ForkException* ex)
        {
            ex->SetActionInstance(this);
            delete ex;
            return false;
        }
    }

    // deposit tokens to postcond
    list<PN_Condition*>::iterator it_var, it_end;
    it_var = this->postconds.begin();
    it_end = this->postconds.end();

    PN_JoinException *join = NULL;

    for(; it_var != it_end; it_var++)
    {
        try
        {
            (*it_var)->DoAsPostcond(this->actionNamespace);
        }
        catch (PN_JoinException *ex)
        {
            if(join == NULL)
                join = ex;
            else
                throw new PN_Error_Exception("Error: multiple calls of Join message (thrown by return place)");
        }
    }

    //DisposeBindings(this->actionNamespace);

    if (join != NULL)
        throw join;
    return true;
}

/** @brief CheckGuard
  *
  * @todo: document this function
  */
bool PN_Guard::Execute(list<PN_Binding>& actionNamespace, bool simulateOnly)
{
    Token *exprResult;

    list<PN_Expression*>::iterator it_var, it_end;
    it_var = this->expr.begin();
    it_end = this->expr.end();

    while(it_var != it_end)
    {
        try
        {
            exprResult = (*it_var)->ExecuteAction(actionNamespace, simulateOnly);
            it_var++;
        }
        catch (PN_ForkException* ex)
        {
            ex->CancelFork();
            //cout << "guard failed on fork exception :";  ex->Print();
            //cout << endl;
            delete ex;
            return false;
        }
        catch (PN_NotEnoughArguments* ex)
        {
            //cout << "guard failed too few arguments :";  ex->Print();
            //cout << endl;
            delete ex;
            return false;
        }
        catch (PN_CantResolve_Exception* ex)
        {
            //cout << "guard failed on resolving :";  ex->Print();
            //cout << endl;
            delete ex;
            return false;
        }
        catch (PN_MessagePass_Exception* ex)
        {
            //cout << "guard failed on misunderstood message :";  ex->Print();
            //cout << endl;
            delete ex;
            return false;
        }

    }

    if(exprResult->GetType() == bool_token && ((BoolToken*)exprResult)->value)
    {
        if(this->nextGuard != NULL)
        {
            if(simulateOnly)
            {
                return this->nextGuard->CheckGuard(actionNamespace);
            }
            else
            {
                return this->nextGuard->ExecuteGuard(actionNamespace);
            }
        }
        else
        {
            return true;
        }
    }
    else
    {
        //cout << "guard failed on bad result; it's either not type BoolToken or particular value is false.";
        return false;
    }
}

/** @brief ExecuteGuard
  *
  * @todo: document this function
  */
bool PN_Guard::ExecuteGuard(list<PN_Binding>& actionNamespace)
{
    //cout << "Executing guard with namespace: ";
    //PrintBindings(actionNamespace);
    return this->Execute(actionNamespace, false);
}

/** @brief CheckGuard
  *
  * @todo: document this function
  */
bool PN_Guard::CheckGuard(list<PN_Binding>& actionNamespace)
{
    //cout << "Trying to check guard with namespace: ";
    //PrintBindings(actionNamespace);
    return this->Execute(actionNamespace, true);
}


/** @brief ExecuteAction
  *
  * @todo: document this function
  */
Token* PN_MessageUnary::ExecuteAction(Token *op, list<PN_Binding>& binds, bool simulateOnly)
{
    list<PN_Operand*> ops;
    //cout<< "sending message \"" << this->text << "\" to token ";
    //op->Print(); cout << endl;
    Token *res;
    if(simulateOnly)
    {
        res = op->SimulateMessage(this->text, ops, binds);
    }
    else
    {
        res = op->SendMessage(this->text, ops, binds);
    }
    return res;
}


/** @brief ExecuteAction
  *
  * @todo: document this function
  */
Token* PN_MessageBin::ExecuteAction(Token *op, list<PN_Binding>& binds, bool simulateOnly)
{
    list<PN_Operand*> ops;
    ops.push_back(this->operand);

    Token *res;
    if(simulateOnly)
    {
        res = op->SimulateMessage(this->text, ops, binds);
    }
    else
    {
        res = op->SendMessage(this->text, ops, binds);
    }

    return res;
}

/** @brief ExecuteAction
  *
  * @todo: document this function
  */
Token* PN_MessageKeys::ExecuteAction(Token *op, list<PN_Binding>& binds, bool simulateOnly)
{
    Token *res;
    if(simulateOnly)
    {
        res = op->SimulateMessage(this->text, this->keys, binds);
    }
    else
    {
        res = op->SendMessage(this->text, this->keys, binds);
    }
    return res;
}

/** @brief ExecuteAction
  *
  * @todo: document this function
  */
Token* PN_Expression::ExecuteAction(list<PN_Binding>& binds, bool simulateOnly)
{
    if (this->rootOperand == NULL)
        throw string("root operand not set\n");
    Token *operand = this->rootOperand->Resolve(binds);

    Token *result;
    if (this->msg != NULL)
        result = this->msg->ExecuteAction(operand, binds, simulateOnly);
    else
        result = operand->Copy();

    //PrintBindings(binds);
    this->UpdateNamespace(binds, result);
    //PrintBindings(binds);
    return result;
}

void PN_Expression::UpdateNamespace(list<PN_Binding>& args, Token* value)
{
    list<string>::iterator it_var, it_end;
    it_var = this->resultVarname.begin();
    it_end = this->resultVarname.end();

    for (; it_var != it_end; it_var++)
    {
        list<PN_Binding>::iterator it_var2, it_end2;
        it_var2 = args.begin();
        it_end2 = args.end();

        bool found = false;

        for (; it_var2 != it_end2; it_var2++)
        {
            if( (*it_var) == it_var2->varName)
            {
                Token *original = it_var2->target;
                delete original;
                original = NULL;
                it_var2->target = value;

                //cout << "update variable '" << *it_var <<"' with value";
                //value -> Print();
                //cout << endl;

                found = true;
                break;
            }
        }
        if(!found)
        {
            //cout << "new variable '" << *it_var <<"' with value";
            //value -> Print();
            //cout << endl;
            args.push_back(PN_Binding(*it_var, value));
        }
    }
}

/** @brief Searches provided var bindings and returns Token which is binded to var name same as this instance
  *
  * @todo: document this function
  */
Token* PN_Operand_Var::Resolve(list<PN_Binding> args)
{
    Token *result;

    bool found = false;
    list<PN_Binding>::iterator it_var, it_end;
    it_var = args.begin();
    it_end = args.end();

    for(; it_var != it_end; it_var++)
    {
        if(it_var->varName == this->varName)
        {
            result = it_var->target;
            found = true;
            break;
        }
    }
    if (!found)
    {
        throw new PN_CantResolve_Exception(this->varName);
    }

    return result;

}

/** @brief Resolve
  *
  * @todo: document this function
  */
Token* PN_Operand_Token::Resolve(list<PN_Binding> args)
{
    return this->bindedToken;
}







