#include "PetriNet.h"
#include "PNsimulation.h"

#include <iostream>

PN::PN()
{
    //ctor
}

PN::~PN()
{
    //cout << "disposing PetriNet \n";
    /// let every transition which belongs to net instance die
    list<PN_Transition*>::iterator it_var, it_end;
    it_var = this->transitions.begin();
    it_end = this->transitions.end();

    for(; it_var != it_end; it_var++)
    {
        (*it_var)->Die();
    }
    ///let simulation system remove transitions, which have just died
    PNsimulation::Instance()->ClearDeadTrans();
}

/** @brief SimulateMessage
  *
  * Default parser for message simulation.
  *
  * @exception PN_MessageNotUnderstood_Exception thrown when message is not handled
  */
Token* PN_Class::SimulateMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    throw new PN_MessageNotUnderstood_Exception(this->AsText(), message);
}

/** @brief SendMessage
  *
  * @todo: document this function
  *
  * @exception PN_MessageNotUnderstood_Exception thrown when message is not handled
  */
Token* PN_Class::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    throw new PN_MessageNotUnderstood_Exception(this->AsText(), message);
}



/** @brief RegisterTransition
  *
  * @todo: document this function
  */
void PN::RegisterTransition(PN_Transition *t)
{
    this->transitions.push_back(t);
    PNsimulation::Instance()->AddTransitions(t);
}

struct NamesDictionary
{
    NamesDictionary(string s, string t)
    {
        source = s;
        target = t;
    }
    string source;
    string target;
};

PN_SyncPort::PN_SyncPort(list<string> args, string name)
{
    this->name = name;
    this->guard = NULL;
    this->argNames = args;
    //int arguments = argNames.size();
    //int parameters = args.size();
}

bool PN_SyncPort::Execute(list<PN_Operand*> ops, list<PN_Binding>& binds)
{

    list<PN_Operand*>::iterator it_op_var, it_op_end;
    it_op_var = ops.begin();
    it_op_end = ops.end();

    list<string>::iterator it_var_aName, it_end_aName;
    it_var_aName = argNames.begin();

    list<PN_Binding> syncNamespace;

    for(; it_op_var!= it_op_end; it_var_aName++, it_op_var++)
    {
        /// register resolvable operand in local namespace (operand is either Token or var contained in action Namespace)
        Token* param = (*it_op_var)->Resolve(binds);
        syncNamespace.push_back(PN_Binding(*it_var_aName, param));
    }

//    PrintBindings(syncNamespace);
     // pop tokens from preconds
    list<PN_Condition*>::iterator it_var, it_end;
    it_var = this->preconds.begin();
    it_end = this->preconds.end();


    for(; it_var != it_end; it_var++)
    {
        (*it_var)->DoAsPrecond(syncNamespace);
    }

    // deposit tokens to postcond
    it_var = this->postconds.begin();
    it_end = this->postconds.end();

    for(; it_var != it_end; it_var++)
    {
        (*it_var)->DoAsPostcond(syncNamespace);
    }
    return true;

}

/** @brief TryBind
  *
  * @todo: document this function
  */
bool PN_SyncPort::TryBind(list<PN_Operand*> ops, list<PN_Binding>& binds)
{
    if(ops.size()!= argNames.size())
    {
        //int operands = ops.size();
        //int arguments = argNames.size();
        throw PN_Error_Exception("Sync port: number of provided operands doesn't match number of requested arguments.");
    }

    list<PN_Operand*>::iterator it_var, it_end;
    it_var = ops.begin();
    it_end = ops.end();

    list<string>::iterator it_var_aName, it_end_aName;
    it_var_aName = argNames.begin();

    list<PN_Binding> syncNamespace;
    list<NamesDictionary> unbindedVars;         /// vars expected to be bind within sync port

    for(; it_var!= it_end; it_var_aName++, it_var++)
    {
        try
        {
            /// register resolvable operand in local namespace (operand is either Token or var contained in action Namespace)
            Token* param = (*it_var)->Resolve(binds);
            syncNamespace.push_back(PN_Binding(*it_var_aName, param));
        }
        catch (PN_CantResolve_Exception* ex)
        {
            /// if operand couldn't be resolved, sync port must try to bind it
            unbindedVars.push_back(NamesDictionary(*it_var_aName, ex->var));
        }
    }

    /// we treat sync port as transition here
    PN_TransitionBinding *syncBind = new PN_TransitionBinding();
    syncBind->bindings = syncNamespace;
    PN_TransitionBinding *bindResult;


/// do binding itself ------------------------------------------------------------
    try
    {
//        cout << "get binds from sync " << name << endl;
        bindResult = this->GetBinding(syncBind);
    }
    catch (PN_Error_Exception ex)
    {
        delete syncBind;
        return false;
    }

    if(bindResult == NULL)
    {
        delete syncBind;
        return false;
    }
    /// try to feedback binded vars to namespace of caller

    list<NamesDictionary>::iterator it_var_d, it_end_d;
    it_var_d = unbindedVars.begin();
    it_end_d = unbindedVars.end();

    /// all unbinded vars must be binded
    bool allBindOK = true;

    /// foreach unbinded var
    for(;it_var_d != it_end_d; it_var_d++)
    {
        list<PN_Binding>::iterator it_var_b, it_end_b;
        it_var_b = bindResult->bindings.begin();
        it_end_b = bindResult->bindings.end();

        bool bindOK = false;

        /// find port binding wich corresponds to actual unbided var
        for (; it_var_b != it_end_b; it_var_b++)
        {
            if(it_var_b->varName == it_var_d->source)
            {
                binds.push_back(PN_Binding(it_var_d->target, it_var_b->target));
                bindOK = true;
                break;
            }
        }

        /// sync port binding failed
        if (!bindOK)
        {
            allBindOK = false;
            break;
        }
    }
    return allBindOK;
}



/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* PNToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    //cout << "PNToken: message \"" << message << "\" accepted\n";
    Token* result;
    try
    {
        result = this->value->SendMessage(message, args, binds);
    }
    catch (PN_MessagePass_Exception ex)
    {
        result = Token::SendMessage(message, args, binds);
    }

    return result;
}

/** @brief SimulateMessage
  *
  * @todo: document this function
  */
Token* PNToken::SimulateMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token* result;
    try
    {
        result = this->value->SimulateMessage(message, args, binds);
    }
    catch (PN_MessagePass_Exception ex)
    {
        result = Token::SimulateMessage(message, args, binds);
    }

    return result;
}




/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* PN_MetaClass::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    if(message == "new")
    {
        return this->GetNewInstance();
    }
    else
    {
        throw new PN_MessageNotUnderstood_Exception (this->AsText(), message);
    }
}

/** @brief PushTokens
  *
  * @todo: document this function
  */
void PN_Place_Return::PushTokens(list<Token*> *tokens)
{
    /*
    cout << "return from method with value ";
    tokens->front()->Print();
    cout << endl;
    */
    throw new PN_JoinMethodException(parent->caller, tokens->front(), this->parent);
}



