#include <typeinfo>
#include "Token.h"
#include <iostream>
#include <cstdint>

Token* FindBinding(list<PN_Binding> *partialBinding, string varName)
{
    list<PN_Binding>::iterator it_var, it_end;

    it_var = partialBinding->begin();
    it_end = partialBinding->end();
    while(it_var != it_end)
    {
        if(varName == it_var->varName)
        {
            return it_var->target;
        }
        it_var++;
    }
    return NULL;
}

PN_Binding::PN_Binding (const PN_Binding& obj)
{
//    cout << "var Name: " << obj.varName << ", target: \n";
//    obj.target->Print();
//    cout << endl;
    this->varName = obj.varName;
    this->target = obj.target;
}

Token::Token()
{
    //ctor
}

Token::~Token()
{
    //dtor
}

/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* Token::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token *result = NULL;

    ///-------------------------Unary messages--------------------------

    if(message == "print")
    {
        cout << this->AsText();
        return new NilToken();
    }

    if(message == "printString")
    {
        return new StringToken(this->AsText());
    }

    ///--------------------------Binary messages-----------------------

    if( message == "==")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        try
        {
            arg1 = args.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        args.pop_front();
        result = new BoolToken(this == arg1);
    }
    else if( message == "~==")
    {
        if (args.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        Token *arg1 = NULL;
        try
        {
            arg1 = args.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        args.pop_front();
        result = new BoolToken(this != arg1);
    }
    else
    {
        throw new PN_MessageNotUnderstood_Exception(this->AsText(), message);
    }

    return result;
}

/** @brief SimulateMessage
  *
  * @todo: document this function
  */
Token* Token::SimulateMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    return this->SendMessage(message, args, binds);
}



/** @brief Compare
  *
  * @todo: document this function
  */
int Token::Compare(Token *other)
{
    if (other == NULL)
    {
        return 1;
    }
    else
    {
        return (this->GetIntertokenOrder()) - (other->GetIntertokenOrder());
    }
}

bool operator < (Token& lhs, Token& rhs)
{
    return (lhs.Compare(&rhs) < 0);
}

bool operator == (Token& lhs, Token& rhs)
{
    return (lhs.Compare(&rhs) == 0);
}

bool compare_tokens_lt (Token* lhs, Token* rhs)
{
    return ((lhs->Compare(rhs)) < 0);
}

int compare_tokens_cmp (Token* lhs, Token* rhs)
{
    if (lhs != NULL)
        return lhs->Compare(rhs);
    else return (rhs == NULL)? 0 : -1;
}

///-------------------------------------------------------------Int Token---------------------------------------------

/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* IntToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token *result = NULL;

    if(message == "print")
    {
        cout << this->value << endl;
        return new NilToken();
    }

///--------------------------Binary messages-----------------------

    if( message == "+")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new IntToken( this->value + ((IntToken*)arg1)->value);
        }
    }
    else if( message == "-")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new IntToken( this->value - ((IntToken*)arg1)->value);
        }
    }
    else if( message == "*")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new IntToken( this->value * ((IntToken*)arg1)->value);
        }
    }
    else if( message == "/")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new IntToken( this->value / ((IntToken*)arg1)->value);
        }
    }
    else if (message == "=")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new BoolToken (this->value == ((IntToken*)arg1)->value);
        }
    }
    else if (message == "!=")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new BoolToken (this->value != ((IntToken*)arg1)->value);
        }
    }
    else if (message == ">")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new BoolToken (this->value > ((IntToken*)arg1)->value);
        }
    }
    else if (message == "<")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new BoolToken (this->value < ((IntToken*)arg1)->value);
        }
    }
    else if (message == ">=")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();
        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new BoolToken (this->value >= ((IntToken*)arg1)->value);
        }
    }
    else if (message == "<=")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception ex)
        {
            throw ex;
        }
        argsLocal.pop_front();

        if(typeid(*arg1) == typeid(IntToken))
        {
            result = new BoolToken (this->value <= ((IntToken*)arg1)->value);
        }
    }

    if (result != NULL)
        return result;
/// ---------------------- Try base class messages -----------------------------

    return Token::SendMessage(message, args, binds);

/// ---------------------- Nothing worked => Token doesn't understand message---
}

/** @brief Compare
  *
  * @todo: document this function
  */
int IntToken::Compare(Token *other)
{
    int relative = this->Token::Compare(other);
    if (relative == 0)
    {
        return this->value - ((IntToken*)other)->value;
    }
    else return relative;
}

/** @brief IntToken
  *
  * @todo: document this function
  */
IntToken::IntToken(int _value)
{
    this->value = _value;
}

///-------------------------------------------------------------Bool Token---------------------------------------------

/** @brief Compare
  *
  * @todo: document this function
  */
int BoolToken::Compare(Token *other)
{
    int relative = this->Token::Compare(other);
    if (relative == 0)
    {
        return this->value - ((IntToken*)other)->value;
    }
    else return relative;
}

/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* BoolToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token *result = NULL;

   ///-------------------------Unary messages--------------------------

    if(message == "print")
    {
        cout << this->value << endl;
        result = new NilToken();
    }

    if (result != NULL)
        return result;

///--------------------------Binary messages-----------------------


    if (message == "=")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception *ex)
        {
            throw ex;
        }
        argsLocal.pop_front();

        if(typeid(*arg1) == typeid(BoolToken))
        {
            result = new BoolToken (this->value == ((BoolToken*)arg1)->value);
        }
    }
    else if (message == "~=")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception *ex)
        {
            throw ex;
        }
        argsLocal.pop_front();

        if(typeid(*arg1) == typeid(BoolToken))
        {
            result = new BoolToken (this->value != ((BoolToken*)arg1)->value);
        }
    }

    if (result != NULL)
        return result;
/// ---------------------- Try base class messages -----------------------------

    return Token::SendMessage(message, args, binds);
}

///-------------------------------------------------------------String Token---------------------------------------------

/** @brief Compare
  *
  * @todo: document this function
  */
int StringToken::Compare(Token *other)
{
    int relative = this->Token::Compare(other);
    if (relative == 0)
    {
        return this->value.compare(((StringToken*)other)->value);
    }
    else return relative;
}

/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* StringToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token *result = NULL;

    ///-------------------------Unary messages--------------------------

    if(message == "print")
    {
        cout << this->value << endl;
        result = new NilToken();
    }

    if (result != NULL)
        return result;

///--------------------------Binary messages-----------------------

    if (message == "=")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception *ex)
        {
            throw ex;
        }
        argsLocal.pop_front();

        if(typeid(*arg1) == typeid(StringToken))
        {
            result = new BoolToken (this->value.compare(((StringToken*)arg1)->value) == 0);
        }
    }

    if (message == ",")
    {
        list<PN_Operand*> argsLocal(args);
        Token *arg1 = NULL;

        if (argsLocal.size() < 1)
            throw PN_Error_Exception("Not enough arguments passed with message.");

        try
        {
            arg1 = argsLocal.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception *ex)
        {
            throw ex;
        }
        argsLocal.pop_front();

        if(typeid(*arg1) == typeid(StringToken))
        {
            result = new StringToken (this->value + ((StringToken*)arg1)->value);
        }
    }

    if (result != NULL)
        return result;
/// ---------------------- Try base class messages -----------------------------

    return Token::SendMessage(message, args, binds);
}

///-------------------------------------------------------------Symbol Token---------------------------------------------

/** @brief Compare
  *
  * @todo: document this function
  */
int SymbolToken::Compare(Token *other)
{
    int relative = this->Token::Compare(other);
    if (relative == 0)
    {
        return this->value.compare(((SymbolToken*)other)->value);
    }
    else return relative;
}

/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* SymbolToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token *result = NULL;

   ///-------------------------Unary messages--------------------------

    if(message == "print")
    {
        cout << this->value << endl;
        result = new NilToken();
    }

    if (result != NULL)
        return result;

///--------------------------Binary messages-----------------------

    if (message == "=")
    {
        if (args.size() < 1)
        {
            //cout << "Not enough arguments passed with message: " << message << endl;
            throw PN_Error_Exception("Not enough arguments passed with message.");
        }

        Token *arg1 = NULL;
        try
        {
            arg1 = args.front()->Resolve(binds);
        }
        catch(PN_CantResolve_Exception *ex)
        {
            throw ex;
        }
        args.pop_front();

        if(typeid(*arg1) == typeid(SymbolToken))
        {
            result = new BoolToken (this->value.compare(((SymbolToken*)arg1)->value) == 0);
        }
    }

    if (result != NULL)
        return result;
/// ---------------------- Try base class messages -----------------------------

    return Token::SendMessage(message, args, binds);
}

int ReferenceToken::Compare(Token *other)
{
    int relative = this->Token::Compare(other);
    if (relative == 0)
    {
        return (intptr_t)this->GetValue() - (intptr_t)((ReferenceToken*)other)->GetValue();
    }
    else return relative;
}

///-------------------------------------------------------------Char Token---------------------------------------------

/** @brief Compare
  *
  * @todo: document this function
  */
int CharToken::Compare(Token *other)
{
    int relative = this->Token::Compare(other);
    if (relative == 0)
    {
        return this->value - ((CharToken*)other)->value;
    }
    else return relative;
}

/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* CharToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token *result = NULL;

    ///-------------------------Unary messages--------------------------

    if(message == "print")
    {
        cout << this->value << endl;
        result = new NilToken();
    }

    if (result != NULL)
        return result;

///--------------------------Binary messages-----------------------

    if (args.size() < 1)
        throw PN_Error_Exception("Not enough arguments passed with message.");

    list<PN_Operand*> argsLocal(args);
    Token *arg1 = NULL;
    try
    {
        arg1 = argsLocal.front()->Resolve(binds);
    }
    catch(PN_CantResolve_Exception *ex)
    {
        throw ex;
    }
    argsLocal.pop_front();

    if (message == "=")
    {
        if(typeid(*arg1) == typeid(StringToken))
        {
            result = new BoolToken (this->value == ((CharToken*)arg1)->value);
        }
    }

    if (result != NULL)
        return result;
/// ---------------------- Try base class messages -----------------------------

    return Token::SendMessage(message, args, binds);
}

///-------------------------------------------------------------List Token---------------------------------------------

string ListToken::AsText()
{
    ostringstream text;
    text << "(";

    list<Token*>::iterator it_var, it_end;
    it_var = this->collection.begin();
    it_end = this->collection.end();

    if(it_var != it_end)
    {
        text << (*it_var)->AsText();
        it_var++;
    }

    for(; it_var != it_end; it_var++)
    {
        text << ", " << (*it_var)->AsText();
    }

    text << ")";
    return text.str();
}

int CompareListTokens (ListToken* lhs, ListToken* rhs)
{
    if(lhs->collection.size() != rhs->collection.size())
        return lhs->collection.size() - rhs->collection.size();
    else
        return compare_lists_cmp(lhs->collection, rhs->collection);
}

int compare_lists_cmp (list<Token*>& lhs, list<Token*>& rhs)
{
    list<Token*>::iterator first1, first2, last1, last2;
    first1 = lhs.begin();
    first2 = rhs.begin();
    last1 = lhs.end();
    last2 = rhs.end();

    while (first1!=last1)
    {
        int cmp = compare_tokens_cmp(*first1, *first2);
        if (first2==last2 || cmp > 0)
            return 1;
        else if (cmp < 0)
            return -1;
        ++first1; ++first2;
    }
    return (first2==last2) ? 0 : -1;
}

/** @brief Compare
  *
  * @todo: document this function
  */
int ListToken::Compare(Token *other)
{
    int relative = this->Token::Compare(other);
    if (relative == 0)
    {
        return CompareListTokens(this, (ListToken*)other);
    }
    else return relative;
}

/** @brief SendMessage
  *
  * @todo: document this function
  */
Token* ListToken::SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
{
    Token* result = NULL;
    if(message == "print")
    {
        cout << this->AsText() << endl;
        result = new NilToken();
    }

    if (result != NULL)
        return result;
/// ---------------------- Try base class messages -----------------------------

    return Token::SendMessage(message, args, binds);

}


