#include <algorithm>

#include "PNparse.h"

#include "PNparse_condition.h"

pnClassContent::~pnClassContent()
{
    list<pnClassContentHolder*>::iterator it_var, it_end;
    it_var = contentList.begin();
    it_end = contentList.end();

    for(; it_var != it_end; it_var++)
    {
        delete *it_var;
    }
}

pnClass::pnClass(ClassHead *head, pnNet *_objectNet )
{
    id = head->name;
    parentName = head->parent;

    if(_objectNet != NULL)
    {
        objectNet = _objectNet;
    }
    delete head;
}

void pnClass::InitContent(pnClassContent *content)
{
    list<pnClassContentHolder*>::iterator it_var, it_end;
    it_var = content->contentList.begin();
    it_end = content->contentList.end();

    for(; it_var != it_end; it_var++)
    {
        switch ((*it_var)->type)
        {
        case method:
            this->addMethod((pnClassMethod*)((*it_var)->contRef));
            break;
        case constructor:
            this->addConstructor((pnClassMethod*)((*it_var)->contRef));
            break;
        case port:
            this->addSyncPort((pnClassSync*)((*it_var)->contRef));
            break;
        default:
            break;
        }
    }
}


pnTransition::pnTransition(string id, pnCondition *_precond, pnCondition *_condition, pnGuard *_guard, pnAction *_action, pnCondition *_postcond)
{
    name = id;
    precond = _precond;
    condition = _condition;
    postcond = _postcond;
    action = _action;
    guard = _guard;
}

/** @brief CheckCorrectness
  *
  * @todo: document this function
  */
void pnNet::CheckCorrectness()
{
    list<pnTransition*>::iterator it_var2, it_end2;
    it_var2 = transitions.begin();
    it_end2 = transitions.end();

    for(; it_var2 != it_end2; it_var2++)
    {
        if(! (*it_var2)->CheckCorrectness(this->places))
        {
            cout << "transition not vallid\n";
        }
    }
}

list<string> pnNet::GetMetaClasses()
{
    list<string> result;
    list<pnTransition*>::iterator it_var, it_end;
    it_var = transitions.begin();
    it_end = transitions.end();

    for(; it_var != it_end; it_var++)
    {
        list<string> metaClasses = (*it_var)->GetMetaClasses();
        result.merge(metaClasses);
    }
    return result;
}

void pnLabel::AddOperand(string key, string varName)
{
    this->messageOriginal += key;
    size_t pos = key.find(':');
    while (pos != string::npos)
    {
        key[pos] = '_';
        pos = key.find(':', pos);
    }
    this->messageTarget += key;
    this->operands.push_back(varName);
}

pnClassSync::pnClassSync(pnLabel *l, pnCondition *_precond, pnCondition *_condition, pnGuard *_guard, pnCondition *_postcond)
{
    this->label = l;
    this->precond = _precond;
    this->condition = _condition;
    this->postcond = _postcond;
    this->guard = _guard;
}

/** @brief CheckCorrectness
  *
  * @todo: document this function
  */
bool pnTransition::CheckCorrectness(list<pnPlace*> placesCol)
{
    return true;
}

void pnTransition::MarkUncheachablePlaces(list<pnPlace*> placesCol)
{
    cout << "transition : " << this->name << endl;
    if (this->precond != NULL)
    {
        cout << "preconds : \n";
        this->precond->MarkUnreachable(placesCol);
    }
    if (this->condition != NULL)
    {
        cout << "conds : \n";
        this->condition->MarkUnreachable(placesCol);
    }
    if (this->postcond != NULL)
    {
        cout << "postconds : \n";
        this->postcond->MarkUnreachable(placesCol);
    }
}

void pnClass::Print()
{
    cout << "class " << id << ", parent: " << parentName << endl;
    cout << "\t object net :";
    if(objectNet != NULL)
    {
        cout << endl;
        objectNet->Print();
    }
    else
        cout << " none \n";
}



void pnNet::Print()
{
    cout << "\t\tplaces:\n";
    list<pnPlace*>::iterator it_var, it_end;
    it_var = places.begin();
    it_end = places.end();

    for(; it_var != it_end; it_var++)
    {
        (*it_var)->Print();
    }
    //cout << endl;

    cout << "\t\ttransitions:\n";
    list<pnTransition*>::iterator it_var2, it_end2;
    it_var2 = transitions.begin();
    it_end2 = transitions.end();

    for(; it_var2 != it_end2; it_var2++)
    {
        (*it_var2)->Print();
    }
    //cout << endl;
}

void pnTransition::Print()
{
    cout << "\t\t    " << name << endl;
    if(this->precond != NULL)
    {
        cout << "\t\t\tprecond: ";
        this->precond->Print();
        cout << endl;
    }
    if(this->condition != NULL)
    {
        cout << "\t\t\tcond:";
        this->condition->Print();
        cout << endl;
    }
    if(this->action != NULL)
    {
        cout << "\t\t\taction: ";
        this->action->Print();
    }
    if(this->postcond != NULL)
    {
        cout << "\t\t\tpostcond";
        this->postcond->Print();
        cout << endl;
    }
}

pnClassMethod::pnClassMethod(pnLabel *l, pnNet *net)
{
    label = l;
    methodNet = net;
    net->MarkUncheachablePlaces();
}

void pnNet::MarkUncheachablePlaces()
{
    list<pnTransition*>::iterator it_var_op, it_end_op;
    it_var_op = this->transitions.begin();
    it_end_op = this->transitions.end();

    for(; it_var_op != it_end_op; it_var_op++)
    {
        (*it_var_op)->MarkUncheachablePlaces(this->places);
    }

}

void pnClassMethod::Print()
{
    cout << this->label->messageTarget << "(";
    list<string>::iterator it_var_op, it_end_op;
    it_var_op = this->label->operands.begin();
    it_end_op = this->label->operands.end();

    for(; it_var_op != it_end_op; it_var_op++)
    {
        cout << *it_var_op << ",";
    }
    cout << ")" << endl;

    cout << "\t\tnet:\n";
    this->methodNet->Print();
    //cout << endl;
}

void PN_AST_Base::Print()
{
    list <pnClass *> classes = this->GetClasses();
    list<pnClass*>::iterator it_var, it_end;
    it_var = classes.begin();
    it_end = classes.end();

    cout << "Starting Class: " << startClass << endl;

    for(; it_var != it_end; it_var++)
    {
        (*it_var)->Print();
    }
    cout << endl;
}

bool PN_AST_Base::DoSemantics()
{
    list <pnClass *> classes = this->GetClasses();
    list<string> builtins;
    builtins.push_back("PN");
    builtins.push_back("Block");
    builtins.push_back("OrderedArray");

    list<pnClass*>::iterator it_var, it_end ,it_var2, it_end2;
    it_var = classes.begin();
    it_end = classes.end();

    for(; it_var != it_end; it_var++)
    {
        pnClass *actClass = *it_var;
        if (std::find(builtins.begin(), builtins.end(), actClass->id) != builtins.end())
        {
            cout << "Class " << actClass->id << " is defined as support class, therefore it can't be redefined in PNtalk code.";
            return false;
        }

        if(actClass->parentName == "PN")
        {
            actClass->parentName = "PN_Class";
            continue;
        }

        it_var2 = classes.begin();
        it_end2 = classes.end();

        bool parentDefined = false;
        for(; it_var2 != it_end2; it_var2++)
        {
            if(actClass->parentName == (*it_var2)->id)
            {
                parentDefined = true;
                break;
            }
        }
        if(!parentDefined)
        {
            cout << "Class " << actClass->id << ": parent class " << actClass->parentName << " is not defined or supported.";
            return false;
        }
    }
    return true;
}

