#ifndef PNparse_H
#define PNparse_H

#include <string>
#include <list>
#include <iostream>

#include "PNparse_condition.h"
#include "PNparse_action.h"
#include "GeneratorFactory.h"

using namespace std;

class ClassHead
{
public:
    string name;
    string parent;
    ClassHead(string _name, string _parent)
    {
        name = _name;
        parent = _parent;
    }
};

typedef enum { method, port, constructor } ContentType;

typedef struct
{
    ContentType type;
    int *contRef;
} pnClassContentHolder;

class pnClassMethod;
class pnClassSync;
class pnNet;

class pnClassContent
{
public:
    list<pnClassContentHolder*> contentList;
    ~pnClassContent();
};

class pnClass
{
protected:
    virtual void addMethod(pnClassMethod* net) = 0;
    virtual void addConstructor(pnClassMethod* net) = 0;
    virtual void addSyncPort(pnClassSync* sync) = 0;

public:
    string parentName;
    string id;
    pnNet *objectNet;

    virtual void Print();
    pnClass(ClassHead *head, pnNet *_objectNet);
    void InitContent(pnClassContent *content);
};

class pnGuard
{

public:
    list<pnExprAssignment*> exprsSimple;
    list<pnExprAssignmentTree*> exprList;
    void AddExpr(pnExprAssignmentTree *exprTree)
    {
        if (exprTree!= NULL)
        {
            exprList.push_back(exprTree);
            string nameSpace = "";
            exprTree->Simplify(&exprsSimple, nameSpace);
        }
    }
};

class pnTransition
{
public:
    string name;
    pnCondition *precond;
    pnCondition *condition;
    pnCondition *postcond;
    pnAction *action;
    pnGuard *guard;
    pnTransition(string id, pnCondition *_precond, pnCondition *_condition, pnGuard *_guard, pnAction *_action, pnCondition *_postcond);
    void MarkUncheachablePlaces(list<pnPlace*> placesCol);
    bool CheckCorrectness(list<pnPlace*> placesCol);
    list<string> GetMetaClasses() {return this->action->GetMetaClasses();}
    void Print();
};

class pnNet
{
public:
    list<pnPlace*> places;
    list<pnTransition*> transitions;
    void AddPlace(pnPlace* place)
    {
        places.push_back(place);
    }
    void AddTransition(pnTransition* trans)
    {
        transitions.push_back(trans);
    }
    void MarkUncheachablePlaces();
    void CheckCorrectness();
    void Print();
    list<string> GetMetaClasses();
};



class pnLabel
{
public:
    string messageOriginal;
    string messageTarget;
    list<string> operands;
    void AddOperand(string key, string varName);
};

class pnClassSync
{
public:
    pnLabel *label;
    pnCondition *precond;
    pnCondition *condition;
    pnCondition *postcond;
    pnGuard *guard;
    pnClassSync(pnLabel *l, pnCondition *_precond, pnCondition *_condition, pnGuard *_guard, pnCondition *_postcond);
};

class pnClassMethod
{
public:
    pnLabel *label;
    pnNet *methodNet;
    pnClassMethod(pnLabel *l, pnNet *net);
    void Print();
};



class PN_AST_Base : public PN_AST
{
protected:
    virtual list<pnClass*> GetClasses() = 0;
public:
    string startClass;
    void Print();

    virtual void SetStartClass(string className) { this->startClass = className;}

    virtual ~PN_AST_Base() {}

    virtual bool DoSemantics();
};

#endif
