#include "PNparse_action.h"
#include <sstream>
#include <typeinfo>

#include "GeneratorFactory.h"

/** @brief Simplify
  *
  * @todo: document this function
  */
void pnExprAssignmentTree::Simplify(list<pnExprAssignment*> *straightList, string nameSpace)
{
    if (this->operand == NULL)
    {
        return;
    }

    pnOperand *op = this->operand;

    string primaryName (nameSpace);
    primaryName.append("__primary__");

    bool simpleOperand = true;
    bool opIsExpr = typeid(*(this->operand)) == typeid(pnOperandExpr);
    if(opIsExpr)
    {
        ((pnOperandExpr*)this->operand)->Simplify(straightList, primaryName, nameSpace);
        op = GeneratorImplementation::Instance()->CreateOperandVar(primaryName);
        simpleOperand = false;
    }

    if(this->msgUnary != NULL)
    {
        this->msgUnary->Simplify(straightList, primaryName, op);
        if (simpleOperand)
        {
            op = GeneratorImplementation::Instance()->CreateOperandVar(primaryName);
            simpleOperand = false;
        }
    }

    if(this->msgBinary != NULL)
    {
        this->msgBinary->Simplify(straightList, primaryName, op, nameSpace);
        if (simpleOperand)
        {
            op = GeneratorImplementation::Instance()->CreateOperandVar(primaryName);
            simpleOperand = false;
        }
    }
    if(this->msgKeys != NULL)
    {
        this->msgKeys->Simplify(straightList, primaryName, op, nameSpace);
        if (simpleOperand)
        {
            op = GeneratorImplementation::Instance()->CreateOperandVar(primaryName);
            simpleOperand = false;
        }
    }

    if (!this->resultVars.empty())
    {
        pnExprAssignment *expr = new pnExprAssignment(op, NULL);
        expr->AddResultVars(this->resultVars);
        straightList->push_back(expr);
    }

}

/** @brief Simplify
  *
  * @todo: document this function
  */
void pnOperandExpr::Simplify(list<pnExprAssignment*> *straightList, string target, string nameSpace)
{
    nameSpace.append("_");
    this->expr->AddResultVar(target);
    this->expr->Simplify(straightList, nameSpace);
}



/** @brief Simplify
  *
  * @todo: document this function
  */
void pnMessageKeys::Simplify(list<pnExprAssignment*> *straightList, string target, pnOperand *reciever, string nameSpace)
{
    list<pnMessageBinary*>::iterator it_var, it_end;
    int counter = 0;
    it_var= this->keys.begin();
    it_end= this->keys.end();

    pnMessageKeys *result = GeneratorImplementation::Instance()->CreateKeysMessage();

    for(; it_var != it_end; it_var++, counter++)
    {
        std::ostringstream argname;
        argname << nameSpace <<"__key" << counter << "__";

        pnMessageBinary *piece = (*it_var)->SimplifyKey(straightList, argname.str(), nameSpace);
        result->AddKey(piece);
    }

    pnExprAssignment *expr = new pnExprAssignment(reciever, result);
    expr->AddResultVar(target);
    straightList->push_back(expr);
}

/** @brief Simplify
  *
  * @todo: document this function
  */
void pnMessageBinary::Simplify(list<pnExprAssignment*> *straightList, string target, pnOperand *reciever, string nameSpace)
{
    pnOperand *arg = this->operand;
    bool simpleOperand = true;
    bool opIsExpr = typeid(*(this->operand)) == typeid(pnOperandExpr);
    if(opIsExpr)
    {
        string argVar (nameSpace);
        argVar.append("__binArg__");

        ((pnOperandExpr*)this->operand)->Simplify(straightList, argVar, nameSpace);

        arg = GeneratorImplementation::Instance()->CreateOperandVar(argVar);
        simpleOperand = false;
        // simplify embedded expr
    }

    if(this->operandChain != NULL)
    {
        string argVar (nameSpace);
        argVar.append("__binArg__");
        this->operandChain->Simplify(straightList, argVar, arg);
        if (simpleOperand)
        {
            arg = GeneratorImplementation::Instance()->CreateOperandVar(argVar);
            simpleOperand = false;
        }
    }

    pnExprAssignment *expr = new pnExprAssignment(reciever, GeneratorImplementation::Instance()->CreateBinaryMessage(this->message, arg, NULL));
    expr->AddResultVar(target);
    straightList->push_back(expr);

    if(this->binChain != NULL)
    {
        this->binChain->Simplify(straightList, target, GeneratorImplementation::Instance()->CreateOperandVar(target), nameSpace);
    }
}

/** @brief Simplify
  *
  * @todo: document this function
  */
pnMessageBinary *pnMessageBinary::SimplifyKey(list<pnExprAssignment*> *straightList, string keyName, string nameSpace)
{
    pnOperand *op = this->operand;

    bool simpleOperand = true;
    bool opIsExpr = typeid(*(this->operand)) == typeid(pnOperandExpr);
    if(opIsExpr)
    {
        ((pnOperandExpr*)this->operand)->Simplify(straightList, keyName, nameSpace);

        op = GeneratorImplementation::Instance()->CreateOperandVar(keyName);
        simpleOperand = false;
    }

    if(this->operandChain != NULL)
    {
        this->operandChain->Simplify(straightList, keyName, op);
        if (simpleOperand)
        {
            op = GeneratorImplementation::Instance()->CreateOperandVar(keyName);
            simpleOperand = false;
        }
    }

    if(this->binChain != NULL)
    {
        this->binChain->Simplify(straightList, keyName, op, nameSpace);
        if (simpleOperand)
        {
            op = GeneratorImplementation::Instance()->CreateOperandVar(keyName);
            simpleOperand = false;
        }
    }

    return GeneratorImplementation::Instance()->CreateBinaryMessage(this->message, op, NULL);
}

/** @brief Simplify
  *
  * @todo: document this function
  */
void pnMessageUnary::Simplify(list<pnExprAssignment*> *straightList, string target, pnOperand *reciever)
{
    pnExprAssignment *expr = new pnExprAssignment(reciever, GeneratorImplementation::Instance()->CreateUnaryMessage(this->message, NULL));
    expr->AddResultVar(target);

    straightList->push_back(expr);

    if(this->unaryChain != NULL)
    {
        this->unaryChain->Simplify(straightList, target, GeneratorImplementation::Instance()->CreateOperandVar(target));
    }
}

list<string> pnExprAssignment::GetMetaClasses()
{
    list<string> result;
    pnOperandVar *varOp = dynamic_cast<pnOperandVar*>(this->target);
    if (varOp != NULL && varOp->IsMetaClass())
    {
        string meta = varOp->GetVarName();
        if(meta == "Time" || meta == "Transcript" || meta == "OrderedCollection")
            result.push_back("PN_Smalltalk");
        else
            result.push_back(meta);
    }
    //TODO search also message bodies?
    return result;
}

list<string> pnAction::GetMetaClasses()
{
    list<string> result;
    list<pnExprAssignment*>::iterator it_var2, it_end2;
    it_var2 = this->exprsSimple.begin();
    it_end2 = this->exprsSimple.end();

    for(; it_var2 != it_end2; it_var2++)
    {
        list<string> metaClasses = (*it_var2)->GetMetaClasses();
        metaClasses.sort();
        result.merge(metaClasses);
    }
    return result;
}

/** @brief Print
  *
  * @todo: document this function
  */
void pnAction::Print()
{
//    list<pnExprAssignmentTree*>::iterator it_var, it_end;
//    it_var= this->exprs.begin();
//    it_end= this->exprs.end();
//
//    for(; it_var != it_end; it_var++)
//    {
//        (*it_var)->Print();
//    }

    list<pnExprAssignment*>::iterator it_var2, it_end2;
    it_var2 = this->exprsSimple.begin();
    it_end2 = this->exprsSimple.end();

    for(; it_var2 != it_end2; it_var2++)
    {
        (*it_var2)->Print();
    }
}

/** @brief Print
  *
  * @todo: document this function
  */
void pnExprAssignmentTree::Print()
{
    list<string>::iterator it_var, it_end;
    it_var = this->resultVars.begin();
    it_end = this->resultVars.end();

    for(; it_var != it_end; it_var++)
    {
        cout << *it_var << " := ";
    }

/// check wheter Expr isn't blank
    if(this->operand != NULL)
        this->operand->Print();
    else return;

    cout << " ";
    if(this->msgUnary!= NULL)
    {
        this->msgUnary->Print();
        cout << " ";
    }
    if(this->msgBinary!= NULL)
    {
        this->msgBinary->Print();
        cout << " ";
    }
    if(this->msgKeys!= NULL)
    {
        this->msgKeys->Print();
        cout << " ";
    }
    cout << "." << endl;
}

/** @brief Print
  *
  * @todo: document this function
  */
void pnExprAssignment::Print()
{
    list<string>::iterator it_var, it_end;
    it_var= this->resultVars.begin();
    it_end= this->resultVars.end();

    for(; it_var != it_end; it_var++)
    {
        cout << *it_var << " := ";
    }

    this->target->Print();
    if(this->msg != NULL)
    {
        cout << " ";
        this->msg->Print();
    }
    cout << endl;
}

/** @brief Print
  *
  * @todo: document this function
  */
void pnMessageKeys::Print()
{
    list<pnMessageBinary*>::iterator it_var, it_end;
    it_var= this->keys.begin();
    it_end= this->keys.end();

    for(; it_var != it_end; it_var++)
    {
        (*it_var)->Print();
    }
}

/** @brief Print
  *
  * @todo: document this function
  */
void pnMessageBinary::Print()
{
    cout << this->message << " ";
    this->operand->Print();
    cout << " ";
    if (this->operandChain != NULL)
    {
        this->operandChain->Print();
    }
    if (this->binChain != NULL)
    {
        this->binChain->Print();
    }
}

/** @brief Print
  *
  * @todo: document this function
  */
void pnMessageUnary::Print()
{
    cout << this->message << " ";

    if (this->unaryChain != NULL)
    {
        this->unaryChain->Print();
    }
}



/** @brief Print
  *
  * @todo: document this function
  */
void pnOperandLiteral::Print()
{
    this->lit->Print();
}

/** @brief Print
  *
  * @todo: document this function
  */
void pnOperandExpr::Print()
{
    cout << "(";
    this->expr->Print();
    cout << ")";
}

/** @brief Print
  *
  * @todo: document this function
  */
void pnOperandVar::Print()
{
    cout << this->varName;
}


