#ifndef PNPARSE_ACTION_H_INCLUDED
#define PNPARSE_ACTION_H_INCLUDED


#include "PNparse_literals.h"
//#include <string>
#include <list>
#include <stdexcept>


using namespace std;

class pnExprAssignment;

class pnOperand
{
public:
    virtual void Print()=0;
    virtual string GenerateNewInstance()=0;
};

class pnMessage
{
protected:
    string message;
public:
    pnMessage(string msg): message(msg) {}
    virtual void Print()=0;
    virtual void GenerateSetAction(ostringstream& text, string actionName, pnOperand *target) = 0;
};

class pnMessageUnary : public pnMessage
{
private:
    pnMessageUnary *unaryChain;
public:
    pnMessageUnary(string msg, pnMessageUnary *nextMsg)
        : pnMessage(msg) , unaryChain(nextMsg) {}
    void Simplify(list<pnExprAssignment*> *straightList, string target, pnOperand *reciever);
    void Print();
};

class pnMessageBinary : public pnMessage
{
private:
    pnMessageUnary *operandChain;
    pnMessageBinary *binChain;
public:
    pnOperand *operand;
    pnMessageBinary(string msg, pnOperand* op, pnMessageUnary *opChain)
        :pnMessage(msg), operandChain(opChain), operand(op)
    {
        binChain = NULL;
    }

    void SetNextMsg(pnMessageBinary *nextMsg)
    {
        binChain = nextMsg;
    }
    void Simplify(list<pnExprAssignment*> *straightList, string target, pnOperand *reciever, string nameSpace);
    pnMessageBinary *SimplifyKey(list<pnExprAssignment*> *straightList, string keyName, string nameSpace);

    string GetMessageName()
    {
        return this->message;
    }
    void Print();
};

class pnMessageKeys : public pnMessage
{
protected:
    list<pnMessageBinary*> keys;
public:
    pnMessageKeys() : pnMessage("") {}
    void AddKey(pnMessageBinary* k)
    {
        keys.push_back(k);
        this->message.append(k->GetMessageName());
    }
    void Simplify(list<pnExprAssignment*> *straightList, string target, pnOperand *reciever, string nameSpace);
    void Print();
};


class pnExprAssignment
{

public:
    list<string> resultVars;
    pnOperand *target;
    pnMessage *msg;

    pnExprAssignment(pnOperand *op, pnMessage *m) : target(op), msg(m) {}
    void AddResultVar(string var)
    {
        resultVars.push_back(var);
    }
    void AddResultVars(list<string> varList)
    {
        resultVars.insert(resultVars.end(), varList.begin(), varList.end());
    }
    list<string> GetMetaClasses();
    void Print();
};

class pnExprAssignmentTree
{
    list<string> resultVars;
    pnOperand *operand;
    pnMessageUnary *msgUnary;
    pnMessageBinary *msgBinary;
    pnMessageKeys *msgKeys;

public:
    pnExprAssignmentTree (pnMessageUnary *unary, pnMessageBinary *binary, pnMessageKeys *keys)
        : msgUnary(unary), msgBinary(binary), msgKeys(keys)
    {
        operand = NULL;
    }
    void SetOperand(pnOperand *op)
    {
        operand = op;
    }
    void AddResultVar(string var)
    {
        resultVars.push_back(var);
    }
    void Simplify(list<pnExprAssignment*> *straightList, string nameSpace);
    void Print();
};

class pnAction
{
    list<pnExprAssignmentTree*> exprs;
public:
    list<pnExprAssignment*> exprsSimple;

    pnAction() { }
    void AddExprTree(pnExprAssignmentTree* t)
    {
        if (t!= NULL)
        {
            exprs.push_back(t);
            string nameSpace = "";
            t->Simplify(&exprsSimple, nameSpace);
        }
    }
    list<string> GetMetaClasses();
    void Print();
};

class pnOperandVar : public pnOperand
{
protected:
    string varName;
    bool isMetaClass;
public:
    pnOperandVar(string name) : varName(name) { isMetaClass = isupper(this->varName.at(0));}
    bool IsMetaClass() {return isMetaClass;}
    string GetVarName() {return varName;}
    void Print();

};
class pnOperandExpr : public pnOperand
{
    pnExprAssignmentTree *expr;
public:
    pnOperandExpr(pnExprAssignmentTree *e) : expr(e)
    {
        if (e == NULL)
            throw runtime_error("blank parentheses - must contain at least operand");
    }
    void Print();
    void Simplify(list<pnExprAssignment*> *straightList, string target, string nameSpace);
    string GenerateNewInstance()
    {
        throw runtime_error("Embedded expression can't be generated directly. Call Simplify() function instead.");
    }
};
class pnOperandLiteral : public pnOperand
{
protected:
    Literal *lit;
public:
    pnOperandLiteral(Literal *l) : lit(l) {}
    void Print();
};

#endif // PNPARSE_ACTION_H_INCLUDED
