#ifndef PNPARSE_MULTISET_H_INCLUDED
#define PNPARSE_MULTISET_H_INCLUDED

#include <list>
#include <sstream>
#include "PNparse_literals.h"

typedef enum { literal, variable, mutisetlist} MultisetListItemType;

class pnMultisetItem;

class pnMultiset
{
public:
    list<pnMultisetItem*> mItems;
    pnMultiset() {}
    ~pnMultiset();
    void AddItem(pnMultisetItem * item);
    void Print();
};

class pnSetItem
{
protected:
    MultisetListItemType type;
public:
    virtual void Print();
    virtual void GenerateAddToMultiSet(ostringstream& text, string varName, string quantity, int level )=0;
// TODO (martin#1#): dodelat kvantitu pomoci promenne
    virtual ~pnSetItem(){}
};

class pnMultisetItem
{
    bool quantityIsRef;
    string refName;
    int quantity;
    pnSetItem *item;
public:
    pnMultisetItem(pnSetItem *_item, int _quantity);
    pnMultisetItem(pnSetItem *_item, string varName);
    ~pnMultisetItem();
    void Print();
    void GenerateAddToMultiSet(ostringstream& text, string varName)
    {
        ostringstream quantString;
        if(quantityIsRef)
        {
            quantString << "\"" << refName << "\"";
        }
        else
        {
            quantString << quantity;
        }
        item->GenerateAddToMultiSet(text, varName, quantString.str(), 1);
// TODO (martin#1#): dodelat kvantitu pomoci promenne
    }
};


class pnMultisetItemList : public pnSetItem
{
protected:
    list<pnSetItem*> mList;
public:
    pnMultisetItemList()
    {
        type = mutisetlist;
    }
    void AddMItem(pnSetItem* item)
    {
        mList.push_back(item);
    }
    ~pnMultisetItemList();
    void Print();
};

class pnMultisetItemVar : public pnSetItem
{
protected:
    string varName;
public:
    pnMultisetItemVar(string name)
    {
        varName = name;
        type = variable;
    }
    virtual void Print()
    {
        cout << varName;
    };
};

class pnMultisetItemLiteral : public pnSetItem
{
protected:
    Literal *_literal;
public:
    pnMultisetItemLiteral(Literal *value)
    {
        _literal = value;
        type = literal;
    }
    virtual ~pnMultisetItemLiteral()
    {
        delete _literal;
    }
    virtual void Print()
    {
        _literal->Print();
    };

};

#endif // PNPARSE_MULTISET_H_INCLUDED
