%start ROOT

%token VARIABLE
%token MAIN CLASS INHERITANCE OBJECT METHOD CONSTRUCTOR
%token SYNC PLACE INIT TRANSITION
%token COND PRECOND POSTCOND GUARD ACTION
%token INTEGER NUMBER DIGIT


%left '+' '-' '*' '/' '>' '<' '&' '|' NE LT GT REFEQ REFNE DIV MOD

%{
#include <stdio.h>
#include "PNparse.h"
#include "CppGenerator.h"
#include <string>
void yyerror(char *);
int yylex(void);
int sym[26];
extern FILE*   yyin;

extern char *yytext;
#define YYDEBUG_LEXER_TEXT yytext

extern int yy_flex_debug;
%}

%union {
int iValue;
char sIndex;
char *strVal;
ClassHead *headPtr;
string *idStr;
pnClassContent *classCont;
pnClassContentHolder *clsContHold;
pnNet *netPtr;
pnPlace *placePtr;
pnTransition *transPtr;
pnCondition *condPtr;
pnMultisetItem *multsetItemPtr;
pnMultiset *multsetPtr;
pnMultisetItemList *multsetListPtr;
pnSetItem * setItemPtr;
Literal *literalPtr;
pnAction *actionPtr;
pnExprAssignmentTree *exprTreePtr;
pnMessageUnary *msgUPtr;
pnMessageBinary *msgBPtr;
pnMessageKeys *msgKPtr;
pnOperand *opPtr;
pnLabel *pnLbl;
pnGuard *pnGrd;
};

%token <iValue> DIGIT INTEGER
%token <strVal> ID KEYSEL STRING_LITERAL
%token <sIndex> SMALLLETTER CAPLETTER BINSELCHAR

%type <headPtr> classhead
%type <classCont> classcontent
%type <idStr> id string symconst symbol unaryMsg binsel keysel
%type <sIndex> charconst char
%type <clsContHold> methodnet constructor sync classcontentitem
%type <netPtr> net objectnet
%type <placePtr> place
%type <transPtr> transition
%type <condPtr> cond precond postcond condS
%type <multsetPtr> multiset
%type <setItemPtr> c
%type <multsetItemPtr> multisetitem
%type <multsetListPtr> list listcontent
%type <literalPtr> literal
%type <actionPtr> action actioncontent expr_assignment_concat
%type <exprTreePtr> expr msgchain expr_assignment
%type <msgUPtr> unaryChain
%type <msgBPtr> binMsg binChain keyMsgSegment
%type <msgKPtr> keyMsg
%type <opPtr> operand
%type <pnLbl> message keys
%type <pnGrd> guard expr_concat


%%

ROOT:
  classes MAIN id classes {GeneratorImplementation::Instance()->SetStartClass(*($3)); }
  ;
classes:
  classes class /* {cout << "classes expand\n";}*/
  | /* {cout << "classes reduce\n";}*/
  ;
class:
  CLASS classhead objectnet classcontent {GeneratorImplementation::Instance()->AddClass($2, $3, $4); delete $4;}
  | CLASS classhead classcontent {GeneratorImplementation::Instance()->AddClass($2, NULL, $3); delete $3;}
  ;
classhead:
  id INHERITANCE id { $$ = new ClassHead(*$1, *$3); delete $1; delete $3; }
  ;
classcontent:
  classcontentitem classcontent {
      $2->contentList.push_back($1);
      $$ = $2; }
  | {$$ = new pnClassContent();}
  ;
classcontentitem:
  methodnet {$$ = $1;}
  |constructor {$$ = $1;}
  |sync {$$ = $1;}
  ;
objectnet:
  OBJECT net {$$ = $2;}
  ;
methodnet:
  METHOD message net {
      pnClassContentHolder *holder = new pnClassContentHolder;
      holder->type = method;
      holder->contRef = (int *)new pnClassMethod($2, $3);
      $$ = holder;}
  ;
constructor:
  CONSTRUCTOR message net {
      pnClassContentHolder *holder = new pnClassContentHolder;
      holder->type = constructor;
      holder->contRef = (int *)new pnClassMethod($2,$3);
      $$ = holder;
    }
  ;
sync:
  SYNC message cond precond guard postcond {
      pnClassContentHolder *holder = new pnClassContentHolder;
      holder->type = port;
      holder->contRef = (int *)new pnClassSync($2, $4, $3, $5, $6);
      $$ = holder;
    }
  ;
message:
  id { pnLabel *lbl = new pnLabel();
        lbl->messageTarget = *$1;
        lbl->messageOriginal = *$1;
        $$ = lbl;
        delete $1; }
  | binsel id {
        pnLabel *lbl = new pnLabel();
        lbl->messageTarget = *$1;
        lbl->messageOriginal = *$1;
        lbl->operands.push_back(*$2);
        $$ = lbl;
        delete $1;
        delete $2; }
  | keys { $$ = $1; }
  ;
keys:
  keys keysel id {
        $1->AddOperand(*$2, *$3);
        $$ = $1;
        delete $2;
        delete $3; }
  |keysel id {
        pnLabel *lbl = new pnLabel();
        lbl->AddOperand(*$1, *$2);
        $$ = lbl;
        delete $1;
        delete $2; }
  ;
net:
  net place { $1->AddPlace($2); $$ = $1; }
  | net transition { $1->AddTransition($2); $$ = $1; }
  | { $$ = new pnNet();}
  ;

place:
  PLACE id'(' multiset ')' init  {$$ = new pnPlace(*$2, $4); delete $2;}
  | PLACE id'(' ')' init  {$$ = new pnPlace(*$2, NULL); delete $2; }
  ;
init:
  INIT '{' actioncontent '}' {;}
  |
  ;
actioncontent:
  expr_assignment_concat {$$ = $1;}
  | temps expr_assignment_concat {$$ = $2;}
  ;
transition:
  TRANSITION id cond precond guard action postcond {$$ = new pnTransition(*$2, $4, $3, $5, $6, $7);delete $2;}
  ;
cond:
  COND condS {$$ = $2;}
  | {$$ = NULL;}
  ;
precond:
  PRECOND condS {$$ = $2;}
  | {$$ = NULL;}
  ;
postcond:
  POSTCOND condS {$$ = $2;}
  | {$$ = NULL;}
  ;
condS:
  id'(' multiset ')' {$$ = new pnCondition(*$1, $3);delete $1;}
  |condS ',' id'(' multiset ')' {$$ = $1; $1->SetNextCond(new pnCondition(*$3, $5), true);}
  ;
guard:
  GUARD '{' expr_concat '}' {$$ = $3;}
  | {$$ = NULL;}
  ;
action:
  ACTION '{' actioncontent '}' {$$ = $3;}
  | {$$ = NULL;}
  ;
multiset:
  multisetitem {pnMultiset *result = new pnMultiset(); result->AddItem($1); $$ = result;}
  | multiset ',' multisetitem { $1->AddItem($3); $$ = $1; }
  ;
multisetitem:
  DIGIT '`' c { $$ = new pnMultisetItem( $3, $1);}
  | INTEGER'`' c { $$ = new pnMultisetItem( $3, $1);}
  | id '`' c { $$ = new pnMultisetItem( $3, *$1);}
  | c { $$ = new pnMultisetItem($1, 1);}
  ;
c:
  literal {$$ = GeneratorImplementation::Instance()->CreateMultisetItemLiteral($1);}
  | id {$$ = GeneratorImplementation::Instance()->CreateMultisetItemVar(*$1); delete $1;}
  | list {$$ = $1;}
  ;
list:
  '(' listcontent  ')' {$$ = $2;}
  ;
listcontent:
  c { pnMultisetItemList *result = GeneratorImplementation::Instance()->CreateMultisetItemList(); result->AddMItem($1); $$ = result;}
  | listcontent ',' c {$1->AddMItem($3); $$ = $1;}
  ;
temps:
  '|' tempscontent '|'
  ;
tempscontent:
  tempscontent id
  |
  ;
operand:
  id {$$ = GeneratorImplementation::Instance()->CreateOperandVar(*$1); delete $1;}
  | literal {$$ = GeneratorImplementation::Instance()->CreateOperandToken($1);}
  | '(' expr_assignment ')'{
      try
      {
	$$ = new pnOperandExpr($2);
      }
      catch (runtime_error ex)
      {
        cerr << ex.what();
      }
    } //TODO chytit vyjimku na prazdny vyraz a volat yyerror
  ;

unaryMsg:
  id {$$ = $1;}
  ;
unaryChain:
  unaryMsg unaryChain {$$ = GeneratorImplementation::Instance()->CreateUnaryMessage(*$1, $2); delete $1; }
  | {$$ = NULL;}
  ;
binMsg:
  binsel operand unaryChain {$$ = GeneratorImplementation::Instance()->CreateBinaryMessage(*$1, $2, $3);}
  ;
binChain:
  binMsg binChain {$1->SetNextMsg($2); $$ = $1;}
  | {$$ = NULL;}
  ;
keysel:
  KEYSEL {$$ = new string($1);}
  ;
keyMsgSegment:
  keysel operand unaryChain binChain {
      $$ = GeneratorImplementation::Instance()->CreateBinaryMessage(*$1, $2, $3); $$->SetNextMsg($4);
      delete $1;
    }
  ;
keyMsg:
  keyMsgSegment {$$ = GeneratorImplementation::Instance()->CreateKeysMessage(); $$->AddKey($1);}
  | keyMsg keyMsgSegment {$$ = $1; $$->AddKey($2);}
  ;
expr_concat:
  expr {$$ = new pnGuard(); $$->AddExpr($1);}
  | expr_concat '.' expr { $$ = $1; $$->AddExpr($3);}
  ;
expr_assignment_concat:
  expr_assignment {$$ = new pnAction(); $$->AddExprTree($1); }
  | expr_assignment_concat '.' expr_assignment {$$ = $1; $1->AddExprTree($3);}
  ;
expr_assignment:
  expr {$$ = $1;}
  | id ':''=' expr_assignment {$4->AddResultVar(*$1); delete $1; $$ = $4;}
  ;
expr:
  operand {$$ = new pnExprAssignmentTree(NULL, NULL, NULL); $$->SetOperand($1);}
  | operand msgchain {$$ = $2; $$->SetOperand($1);}
  | {$$ = NULL;}
  ;
msgchain:
  unaryMsg unaryChain binChain {
        pnMessageUnary *uMsg = GeneratorImplementation::Instance()->CreateUnaryMessage(*$1, $2);
        delete $1;
        $$ = new pnExprAssignmentTree(uMsg, $3, NULL);
    }
  | unaryMsg unaryChain binChain keyMsg {
        pnMessageUnary *uMsg = GeneratorImplementation::Instance()->CreateUnaryMessage(*$1, $2);
        delete $1;
        $$ = new pnExprAssignmentTree(uMsg, $3, $4);
    }
  | binMsg binChain {$1->SetNextMsg($2); $$ = new pnExprAssignmentTree(NULL, $1, NULL);}
  | binMsg binChain keyMsg {$1->SetNextMsg($2); $$ = new pnExprAssignmentTree(NULL, $1, $3);}
  | keyMsg {$$ = new pnExprAssignmentTree(NULL, NULL, $1);}
  ;

binsel:
  BINSELCHAR {$$ = new string(1, $1);}
  | '=' {$$ = new string(1, '=');}
  | ',' {$$ = new string(1, ',');}
  | binsel BINSELCHAR { $1->append(1, $2); $$ = $1;}
  | binsel '=' { $1->append(1, '='); $$ = $1;}
  ;
literal:
  DIGIT {$$ = GeneratorImplementation::Instance()->CreateMultisetItemInt($1);}
  | INTEGER {$$ = GeneratorImplementation::Instance()->CreateMultisetItemInt($1);}
  /* TODO po vyreseni number pridat FLOAT */
  | string {$$ = GeneratorImplementation::Instance()->CreateMultisetItemString(*$1); delete $1;}
  | charconst {$$ = GeneratorImplementation::Instance()->CreateMultisetItemChar($1);}
  | symconst {$$ = GeneratorImplementation::Instance()->CreateMultisetItemSymbol(*$1); delete $1;}
  | arrayconst {$$ = NULL;}
  ;
arrayconst:
  '#' array
  ;
array:
  '(' arraycontent ')'
  ;
arraycontent:
  /*number
  |*/
  arraycontent string
  | arraycontent id
  | arraycontent array
  | arraycontent charconst
  |
  ;
/* number:
  [-][[dig]+ "r"] [hexDig]+ ["." [hexDig]+] ["e"["-"][dig]+].
  ;*/
string:
  STRING_LITERAL {$$ = new string($1);
        $$->erase($$->begin());
        $$->erase($$->end()-1);
        }
  ;
charconst:
  '$' char {$$ = $2;}
  | '$' SMALLLETTER {$$ = $2;}
  | '$' CAPLETTER {$$ = $2;}
  ;
symconst:
  '#'symbol {$$ = $2;}
  ;
symbol:
  id {$$ = $1;}
  | string {$$ = $1;}
  ;
id:
  ID { $$ = new string($1);}
  | SMALLLETTER { $$ = new string(1,$1);}
  | CAPLETTER { $$ = new string(1,$1);}
  ;

/*hexDig:
  DIGIT
  | "A".."F"
  ;*/
letter:
  SMALLLETTER
  | CAPLETTER
  ;
char:
  DIGIT {$$ = $1 + '0';}
  | BINSELCHAR
  | '=' {$$ = '=';}
  | ',' {$$ = ',';}
  | '[' {$$ = '[';}
  | ']' {$$ = ']';}
  | '{' {$$ = '{';}
  | '}' {$$ = '}';}
  | '(' {$$ = '(';}
  | ')' {$$ = ')';}
  | '_' {$$ = '_';}
  | '^' {$$ = '^';}
  | ';' {$$ = ';';}
  | '$' {$$ = '$';}
  | '#' {$$ = '#';}
  | ':' {$$ = ':';}
  | '.' {$$ = '.';}
  | '`' {$$ = '`';}
  ;

%%

void yyerror(char *s) {
fprintf(stderr, "%s\n", s);
return;
}
int main(int argc, char* argv[])
{
    char* in = NULL;
    string dir = "output";
    string programName = "";
    for (int i = 1; i < argc; i++)
    {
        if (i + 1 != argc)
        {
            if (string(argv[i]) == "-dir")
            {
                if (i + 1 != argc)
                {
                    dir = argv[i + 1];
                }
                else cout << "Missing argument: Directory must define argument.\n";
            }
            if (string(argv[i]) == "-in")
            {
                cout << "parsing in";
                if (i + 1 != argc)
                {
                    in = argv[i + 1];
                }
                else cout << "Missing argument: Input file not specified.\n";
            }
            if (string(argv[i]) == "-out")
            {
                if (i + 1 != argc)
                {
                    programName = argv[i + 1];
                }
                else cout << "Missing argument: output program name not specified.\n";
            }
        }
        std::cout << "(i = "<< i <<") "<< argv[i] << ", ";

    }

    if (in == NULL)
    {
        cout << "\nInput file not set. Usage: PNTranslator -in inputfile [-dir directoryOutput] [-out programName]\n";
        return 0;
    }

    if (programName == "")
    {
        programName = dir;
    }

    PN_AST_Base *syntaxTree =  new CppGenerator();
    GeneratorImplementation::Instance()->SetImplementation(syntaxTree);
    yyin = fopen(in, "r");
    yy_flex_debug= 1;
    yydebug = 1;

    cout << "parsing...\n";
    yyparse();

    cout << "Printing...\n";
    syntaxTree->Print();

    if (syntaxTree->DoSemantics())
    {
        syntaxTree->GenerateCode(dir, programName);
    }
    delete syntaxTree;
    return 0;
}
